"""Discretised functional data module.

This module defines a class for representing functional data as a series of
lists of values, each representing the observation of a function measured in a
list of discretisation points.

"""
from __future__ import annotations

import copy
import numbers
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import findiff
import numpy as np
import pandas.api.extensions
import scipy.stats.mstats
from matplotlib.figure import Figure

from .._utils import (
    _check_array_key,
    _int_to_real,
    _to_domain_range,
    _to_grid_points,
    constants,
)
from ._functional_data import FData
from ._typing import (
    ArrayLike,
    DomainRange,
    DomainRangeLike,
    GridPoints,
    GridPointsLike,
    LabelTupleLike,
    NDArrayInt,
)
from .basis import Basis
from .evaluator import Evaluator
from .extrapolation import ExtrapolationLike
from .interpolation import SplineInterpolation

if TYPE_CHECKING:
    from .. import FDataBasis

T = TypeVar("T", bound='FDataGrid')


class FDataGrid(FData):  # noqa: WPS214
    r"""Represent discretised functional data.

    Class for representing functional data as a set of curves discretised
    in a grid of points.

    Attributes:
        data_matrix: a matrix where each entry of the first
            axis contains the values of a functional datum evaluated at the
            points of discretisation.
        grid_points: 2 dimension matrix where each row
            contains the points of dicretisation for each axis of data_matrix.
        domain_range: 2 dimension matrix where each row
            contains the bounds of the interval in which the functional data
            is considered to exist for each one of the axies.
        dataset_name: name of the dataset.
        argument_names: tuple containing the names of the different
            arguments.
        coordinate_names: tuple containing the names of the different
            coordinate functions.
        extrapolation: defines the default type of
            extrapolation. By default None, which does not apply any type of
            extrapolation. See `Extrapolation` for detailled information of the
            types of extrapolation.
        interpolation: Defines the type of interpolation
            applied in `evaluate`.

    Examples:
        Representation of a functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`,
        with 3 discretization points.

        >>> data_matrix = [[1, 2, 3], [4, 5, 6]]
        >>> grid_points = [2, 4, 5]
        >>> FDataGrid(data_matrix, grid_points)
        FDataGrid(
            array([[[ 1.],
                    [ 2.],
                    [ 3.]],
        <BLANKLINE>
                   [[ 4.],
                    [ 5.],
                    [ 6.]]]),
            grid_points=(array([ 2., 4., 5.]),),
            domain_range=((2.0, 5.0),),
            ...)

        The number of columns of data_matrix have to be the length of
        grid_points.

        >>> FDataGrid(np.array([1,2,4,5,8]), range(6))
        Traceback (most recent call last):
            ....
        ValueError: Incorrect dimension in data_matrix and grid_points...


        FDataGrid support higher dimensional data both in the domain and image.
        Representation of a functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}\longmapsto\mathbb{R}^2`.

        >>> data_matrix = [[[1, 0.3], [2, 0.4]], [[2, 0.5], [3, 0.6]]]
        >>> grid_points = [2, 4]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>> fd.dim_domain, fd.dim_codomain
        (1, 2)

        Representation of a functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}^2\longmapsto\mathbb{R}`.

        >>> data_matrix = [[[1, 0.3], [2, 0.4]], [[2, 0.5], [3, 0.6]]]
        >>> grid_points = [[2, 4], [3,6]]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>> fd.dim_domain, fd.dim_codomain
        (2, 1)

    """

    def __init__(  # noqa: WPS211
        self,
        data_matrix: ArrayLike,
        grid_points: Optional[GridPointsLike] = None,
        *,
        sample_points: Optional[GridPointsLike] = None,
        domain_range: Optional[DomainRangeLike] = None,
        dataset_label: Optional[str] = None,
        dataset_name: Optional[str] = None,
        argument_names: Optional[LabelTupleLike] = None,
        coordinate_names: Optional[LabelTupleLike] = None,
        sample_names: Optional[LabelTupleLike] = None,
        axes_labels: Optional[LabelTupleLike] = None,
        extrapolation: Optional[ExtrapolationLike] = None,
        interpolation: Optional[Evaluator] = None,
    ):
        """Construct a FDataGrid object."""
        if sample_points is not None:
            warnings.warn(
                "Parameter sample_points is deprecated. Use the "
                "parameter grid_points instead.",
                DeprecationWarning,
            )
            grid_points = sample_points

        self.data_matrix = _int_to_real(np.atleast_2d(data_matrix))

        if grid_points is None:
            self.grid_points = _to_grid_points([
                np.linspace(0, 1, self.data_matrix.shape[i])
                for i in range(1, self.data_matrix.ndim)
            ])

        else:
            # Check that the dimension of the data matches the grid_points
            # list

            self.grid_points = _to_grid_points(grid_points)

            data_shape = self.data_matrix.shape[1: 1 + self.dim_domain]
            grid_points_shape = [len(i) for i in self.grid_points]

            if not np.array_equal(data_shape, grid_points_shape):
                raise ValueError(
                    f"Incorrect dimension in data_matrix and "
                    f"grid_points. Data has shape {data_shape} and grid "
                    f"points have shape {grid_points_shape}",
                )

        self._sample_range = tuple(
            (s[0], s[-1]) for s in self.grid_points
        )

        if domain_range is None:
            domain_range = self.sample_range
            # Default value for domain_range is a list of tuples with
            # the first and last element of each list of the grid_points.

        self._domain_range = _to_domain_range(domain_range)

        if len(self._domain_range) != self.dim_domain:
            raise ValueError("Incorrect shape of domain_range.")

        for domain_range, grid_points in zip(
            self._domain_range,
            self.grid_points,
        ):
            if (
                domain_range[0] > grid_points[0]
                or domain_range[-1] < grid_points[-1]
            ):
                raise ValueError(
                    "Grid points must be within the domain range.",
                )

        # Adjust the data matrix if the dimension of the image is one
        if self.data_matrix.ndim == 1 + self.dim_domain:
            self.data_matrix = self.data_matrix[..., np.newaxis]

        self.interpolation = interpolation  # type: ignore

        super().__init__(
            extrapolation=extrapolation,
            dataset_label=dataset_label,
            dataset_name=dataset_name,
            axes_labels=axes_labels,
            argument_names=argument_names,
            coordinate_names=coordinate_names,
            sample_names=sample_names,
        )

    def round(  # noqa: WPS125
        self,
        decimals: int = 0,
        out: Optional[FDataGrid] = None,
    ) -> FDataGrid:
        """Evenly round to the given number of decimals.

        .. deprecated:: 0.6
            Use :func:`numpy.round` function instead.

        Args:
            decimals: Number of decimal places to round to.
                If decimals is negative, it specifies the number of
                positions to the left of the decimal point. Defaults to 0.
            out: FDataGrid where to place the result, if any.

        Returns:
            Returns a FDataGrid object where all elements
            in its data_matrix are rounded.

        """
        out_matrix = None if out is None else out.data_matrix

        if (
            out is not None
            and (
                self.domain_range != out.domain_range
                or not all(
                    np.array_equal(a, b)
                    for a, b in zip(self.grid_points, out.grid_points)
                )
                or self.data_matrix.shape != out.data_matrix.shape
            )
        ):
            raise ValueError("out parameter is not valid")

        data_matrix = np.round(
            self.data_matrix,
            decimals=decimals,
            out=out_matrix,
        )

        return self.copy(data_matrix=data_matrix) if out is None else out

    @property
    def sample_points(self) -> GridPoints:
        warnings.warn(
            "Parameter sample_points is deprecated. Use the "
            "parameter grid_points instead.",
            DeprecationWarning,
        )
        return self.grid_points

    @property
    def dim_domain(self) -> int:
        return len(self.grid_points)

    @property
    def dim_codomain(self) -> int:
        try:
            # The dimension of the image is the length of the array that can
            #  be extracted from the data_matrix using all the dimensions of
            #  the domain.
            return self.data_matrix.shape[1 + self.dim_domain]
        # If there is no array that means the dimension of the image is 1.
        except IndexError:
            return 1

    @property
    def coordinates(self: T) -> _CoordinateIterator[T]:
        r"""Returns an object to access to the image coordinates.

        If the functional object contains multivariate samples
        :math:`f: \mathbb{R}^n \rightarrow \mathbb{R}^d`, this class allows
        iterate and get coordinates of the vector
        :math:`f = (f_0, ..., f_{d-1})`.

        Examples:

            We will construct a dataset of curves in :math:`\mathbb{R}^3`

            >>> from skfda.datasets import make_multimodal_samples
            >>> fd = make_multimodal_samples(dim_codomain=3, random_state=0)
            >>> fd.dim_codomain
            3

            The functions of this dataset are vectorial functions
            :math:`f(t) = (f_0(t), f_1(t), f_2(t))`. We can obtain a specific
            component of the vector, for example, the first one.

            >>> fd_0 = fd.coordinates[0]
            >>> fd_0
            FDataGrid(...)

            The object returned has image dimension equal to 1

            >>> fd_0.dim_codomain
            1

            Or we can get multiple components, it can be accesed as a 1-d
            numpy array of coordinates, for example, :math:`(f_0(t), f_1(t))`.

            >>> fd_01 = fd.coordinates[0:2]
            >>> fd_01.dim_codomain
            2

            We can use this method to iterate throught all the coordinates.

            >>> for fd_i in fd.coordinates:
            ...     fd_i.dim_codomain
            1
            1
            1

            This object can be used to split a FDataGrid in a list with
            their components.

            >>> fd_list = list(fd.coordinates)
            >>> len(fd_list)
            3

        """

        return _CoordinateIterator(self)

    @property
    def n_samples(self) -> int:
        """
        Return the number of samples.

        This is also the number of rows of the data_matrix.

        Returns:
            Number of samples of the FDataGrid object.

        """
        return self.data_matrix.shape[0]

    @property
    def sample_range(self) -> DomainRange:
        """
        Return the sample range of the function.

        This contains the minimum and maximum values of the grid points in
        each dimension.

        It does not have to be equal to the `domain_range`.
        """
        return self._sample_range

    @property
    def domain_range(self) -> DomainRange:
        """
        Return the :term:`domain range` of the function.

        It does not have to be equal to the `sample_range`.

        """
        return self._domain_range

    @property
    def interpolation(self) -> Evaluator:
        """Define the type of interpolation applied in `evaluate`."""
        return self._interpolation

    @interpolation.setter
    def interpolation(self, new_interpolation: Optional[Evaluator]) -> None:

        if new_interpolation is None:
            new_interpolation = SplineInterpolation()

        self._interpolation = new_interpolation

    def _evaluate(
        self,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:

        return self.interpolation(  # type: ignore
            self,
            eval_points,
            aligned=aligned,
        )

    def derivative(self: T, *, order: int = 1) -> T:
        """Differentiate a FDataGrid object.

        It is calculated using central finite differences when possible. In
        the extremes, forward and backward finite differences with accuracy
        2 are used.

        Args:
            order: Order of the derivative. Defaults to one.

        Returns:
            Derivative function.

        Examples:
            First order derivative

            >>> fdata = FDataGrid([1,2,4,5,8], range(5))
            >>> fdata.derivative()
            FDataGrid(
                array([[[ 0.5],
                        [ 1.5],
                        [ 1.5],
                        [ 2. ],
                        [ 4. ]]]),
                grid_points=(array([ 0., 1., 2., 3., 4.]),),
                domain_range=((0.0, 4.0),),
                ...)

            Second order derivative

            >>> fdata = FDataGrid([1,2,4,5,8], range(5))
            >>> fdata.derivative(order=2)
            FDataGrid(
                array([[[ 3.],
                        [ 1.],
                        [-1.],
                        [ 2.],
                        [ 5.]]]),
                grid_points=(array([ 0., 1., 2., 3., 4.]),),
                domain_range=((0.0, 4.0),),
                ...)

        """
        order_list = np.atleast_1d(order)
        if order_list.ndim != 1 or len(order_list) != self.dim_domain:
            raise ValueError("The order for each partial should be specified.")

        operator = findiff.FinDiff(*[
            (1 + i, *z)
            for i, z in enumerate(
                zip(self.grid_points, order_list),
            )
        ])
        data_matrix = operator(self.data_matrix.astype(float))

        return self.copy(
            data_matrix=data_matrix,
        )

    def _check_same_dimensions(self: T, other: T) -> None:
        if self.data_matrix.shape[1:-1] != other.data_matrix.shape[1:-1]:
            raise ValueError("Error in columns dimensions")
        if not np.array_equal(self.grid_points, other.grid_points):
            raise ValueError("Grid points for both objects must be equal")

    def sum(  # noqa: WPS125
        self: T,
        *,
        axis: Optional[int] = None,
        out: None = None,
        keepdims: bool = False,
        skipna: bool = False,
        min_count: int = 0,
    ) -> T:
        """Compute the sum of all the samples.

        Args:
            axis: Used for compatibility with numpy. Must be None or 0.
            out: Used for compatibility with numpy. Must be None.
            keepdims: Used for compatibility with numpy. Must be False.
            skipna: Wether the NaNs are ignored or not.
            min_count: Number of valid (non NaN) data to have in order
                for the a variable to not be NaN when `skipna` is
                `True`.

        Returns:
            A FDataGrid object with just one sample representing
            the sum of all the samples in the original object.

        Examples:
            >>> from skfda import FDataGrid
            >>> data_matrix = [[0.5, 1, 2, .5], [1.5, 1, 4, .5]]
            >>> FDataGrid(data_matrix).sum()
            FDataGrid(
                array([[[ 2.],
                        [ 2.],
                        [ 6.],
                        [ 1.]]]),
                ...)

        """
        super().sum(axis=axis, out=out, keepdims=keepdims, skipna=skipna)

        data = (
            np.nansum(self.data_matrix, axis=0, keepdims=True) if skipna
            else np.sum(self.data_matrix, axis=0, keepdims=True)
        )

        if min_count > 0:
            valid = ~np.isnan(self.data_matrix)
            n_valid = np.sum(valid, axis=0)
            data[n_valid < min_count] = np.NaN

        return self.copy(
            data_matrix=data,
            sample_names=(None,),
        )

    def var(self: T) -> T:
        """Compute the variance of a set of samples in a FDataGrid object.

        Returns:
            A FDataGrid object with just one sample representing the
            variance of all the samples in the original FDataGrid object.

        """
        return self.copy(
            data_matrix=np.array([np.var(self.data_matrix, 0)]),
            sample_names=("variance",),
        )

    def cov(self: T) -> T:
        """Compute the covariance.

        Calculates the covariance matrix representing the covariance of the
        functional samples at the observation points.

        Returns:
            Covariance function.

        """
        dataset_name = (
            f"{self.dataset_name} - covariance"
            if self.dataset_name is not None else None
        )

        if self.dim_domain != 1 or self.dim_codomain != 1:
            raise NotImplementedError(
                "Covariance only implemented "
                "for univariate functions",
            )

        return self.copy(
            data_matrix=np.cov(
                self.data_matrix[..., 0],
                rowvar=False,
            )[np.newaxis, ...],
            grid_points=[
                self.grid_points[0],
                self.grid_points[0],
            ],
            domain_range=[
                self.domain_range[0],
                self.domain_range[0],
            ],
            dataset_name=dataset_name,
            argument_names=self.argument_names * 2,
            sample_names=("covariance",),
        )

    def gmean(self: T) -> T:
        """Compute the geometric mean of all samples in the FDataGrid object.

        Returns:
            A FDataGrid object with just one sample representing
            the geometric mean of all the samples in the original
            FDataGrid object.

        """
        return self.copy(
            data_matrix=[
                scipy.stats.mstats.gmean(self.data_matrix, 0),
            ],
            sample_names=("geometric mean",),
        )

    def equals(self, other: object) -> bool:
        """Comparison of FDataGrid objects."""
        if not super().equals(other):
            return False

        other = cast(FDataGrid, other)

        if not np.array_equal(self.data_matrix, other.data_matrix):
            return False

        # Comparison of the domain
        if (
            not np.array_equal(self.domain_range, other.domain_range)
            or len(self.grid_points) != len(other.grid_points)
            or not all(
                np.array_equal(a, b)
                for a, b in zip(self.grid_points, other.grid_points)
            )
        ):
            return False

        return self.interpolation == other.interpolation

    def __eq__(self, other: object) -> np.ndarray:  # type: ignore[override]
        """Elementwise equality of FDataGrid."""
        if not isinstance(other, type(self)) or self.dtype != other.dtype:
            if other is pandas.NA:
                return self.isna()
            if pandas.api.types.is_list_like(other) and not isinstance(
                other, (pandas.Series, pandas.Index, pandas.DataFrame),
            ):
                return np.concatenate([x == y for x, y in zip(self, other)])

            return NotImplemented

        if len(self) != len(other) and len(self) != 1 and len(other) != 1:
            raise ValueError(
                f"Different lengths: "
                f"len(self)={len(self)} and "
                f"len(other)={len(other)}",
            )

        return np.all(
            self.data_matrix == other.data_matrix,
            axis=tuple(range(1, self.data_matrix.ndim)),
        )

    def _get_op_matrix(
        self,
        other: Union[T, np.ndarray, float],
    ) -> Union[None, float, np.ndarray]:
        if isinstance(other, numbers.Real):
            return float(other)
        elif isinstance(other, np.ndarray):

            if other.shape in {(), (1,)}:
                return other
            elif other.shape == (self.n_samples,):
                other_index = (
                    (slice(None),) + (np.newaxis,)
                    * (self.data_matrix.ndim - 1)
                )

                return other[other_index]

        elif isinstance(other, FDataGrid):
            self._check_same_dimensions(other)
            return other.data_matrix

        return None

    def __add__(self: T, other: Union[T, np.ndarray, float]) -> T:

        data_matrix = self._get_op_matrix(other)
        if data_matrix is None:
            return NotImplemented

        return self._copy_op(other, data_matrix=self.data_matrix + data_matrix)

    def __radd__(self: T, other: Union[T, np.ndarray, float]) -> T:

        return self.__add__(other)

    def __sub__(self: T, other: Union[T, np.ndarray, float]) -> T:

        data_matrix = self._get_op_matrix(other)
        if data_matrix is None:
            return NotImplemented

        return self._copy_op(other, data_matrix=self.data_matrix - data_matrix)

    def __rsub__(self: T, other: Union[T, np.ndarray, float]) -> T:

        data_matrix = self._get_op_matrix(other)
        if data_matrix is None:
            return NotImplemented

        return self.copy(data_matrix=data_matrix - self.data_matrix)

    def __mul__(self: T, other: Union[T, np.ndarray, float]) -> T:

        data_matrix = self._get_op_matrix(other)
        if data_matrix is None:
            return NotImplemented

        return self._copy_op(other, data_matrix=self.data_matrix * data_matrix)

    def __rmul__(self: T, other: Union[T, np.ndarray, float]) -> T:

        return self.__mul__(other)

    def __truediv__(self: T, other: Union[T, np.ndarray, float]) -> T:

        data_matrix = self._get_op_matrix(other)
        if data_matrix is None:
            return NotImplemented

        return self._copy_op(other, data_matrix=self.data_matrix / data_matrix)

    def __rtruediv__(self: T, other: Union[T, np.ndarray, float]) -> T:

        data_matrix = self._get_op_matrix(other)
        if data_matrix is None:
            return NotImplemented

        return self._copy_op(other, data_matrix=data_matrix / self.data_matrix)

    def concatenate(self: T, *others: T, as_coordinates: bool = False) -> T:
        """Join samples from a similar FDataGrid object.

        Joins samples from another FDataGrid object if it has the same
        dimensions and sampling points.

        Args:
            others: Objects to be concatenated.
            as_coordinates:  If False concatenates as
                new samples, else, concatenates the other functions as
                new components of the image. Defaults to false.

        Returns:
            FDataGrid object with the samples from the original objects.

        Examples:
            >>> fd = FDataGrid([1,2,4,5,8], range(5))
            >>> fd_2 = FDataGrid([3,4,7,9,2], range(5))
            >>> fd.concatenate(fd_2)
            FDataGrid(
                array([[[ 1.],
                        [ 2.],
                        [ 4.],
                        [ 5.],
                        [ 8.]],
            <BLANKLINE>
                       [[ 3.],
                        [ 4.],
                        [ 7.],
                        [ 9.],
                        [ 2.]]]),
                grid_points=(array([ 0., 1., 2., 3., 4.]),),
                domain_range=((0.0, 4.0),),
                ...)

        """
        # Checks
        if not as_coordinates:
            for other in others:
                self._check_same_dimensions(other)

        elif not all(
            np.array_equal(self.grid_points, other.grid_points)
            for other in others
        ):
            raise ValueError(
                "All the FDataGrids must be sampled in the  same "
                "grid points.",
            )

        elif any(self.n_samples != other.n_samples for other in others):

            raise ValueError(
                f"All the FDataGrids must contain the same "
                f"number of samples {self.n_samples} to "
                f"concatenate as a new coordinate.",
            )

        data = [self.data_matrix] + [other.data_matrix for other in others]

        if as_coordinates:

            coordinate_names = [fd.coordinate_names for fd in (self, *others)]

            return self.copy(
                data_matrix=np.concatenate(data, axis=-1),
                coordinate_names=sum(coordinate_names, ()),
            )

        sample_names = [fd.sample_names for fd in (self, *others)]

        return self.copy(
            data_matrix=np.concatenate(data, axis=0),
            sample_names=sum(sample_names, ()),
        )

    def scatter(self, *args: Any, **kwargs: Any) -> Figure:
        """Scatter plot of the FDatGrid object.

        Args:
            args: positional arguments to be passed to the
                matplotlib.pyplot.scatter function.
            kwargs: keyword arguments to be passed to the
                matplotlib.pyplot.scatter function.

        Returns:
            Figure object in which the graphs are plotted.


        """
        from ..exploratory.visualization.representation import ScatterPlot

        return ScatterPlot(self, *args, **kwargs).plot()

    def to_basis(self, basis: Basis, **kwargs: Any) -> FDataBasis:
        """Return the basis representation of the object.

        Args:
            basis(Basis): basis object in which the functional data are
                going to be represented.
            kwargs: keyword arguments to be passed to
                FDataBasis.from_data().

        Returns:
            FDataBasis: Basis representation of the funtional data
            object.

        Examples:
            >>> import numpy as np
            >>> import skfda
            >>> t = np.linspace(0, 1, 5)
            >>> x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t) + 2
            >>> x
            array([ 3.,  3.,  1.,  1.,  3.])

            >>> fd = FDataGrid(x, t)
            >>> basis = skfda.representation.basis.Fourier(n_basis=3)
            >>> fd_b = fd.to_basis(basis)
            >>> fd_b.coefficients.round(2)
            array([[ 2.  , 0.71, 0.71]])

        """
        from ..preprocessing.smoothing import BasisSmoother

        if self.dim_domain != basis.dim_domain:
            raise ValueError(
                f"The domain of the function has "
                f"dimension {self.dim_domain} "
                f"but the domain of the basis has "
                f"dimension {basis.dim_domain}",
            )
        elif self.dim_codomain != basis.dim_codomain:
            raise ValueError(
                f"The codomain of the function has "
                f"dimension {self.dim_codomain} "
                f"but the codomain of the basis has "
                f"dimension {basis.dim_codomain}",
            )

        # Readjust the domain range if there was not an explicit one
        if not basis.is_domain_range_fixed():
            basis = basis.copy(domain_range=self.domain_range)

        smoother = BasisSmoother(
            basis=basis,
            **kwargs,
            return_basis=True,
        )

        return smoother.fit_transform(self)

    def to_grid(  # noqa: D102
        self: T,
        grid_points: Optional[GridPointsLike] = None,
        *,
        sample_points: Optional[GridPointsLike] = None,
    ) -> T:

        if sample_points is not None:
            warnings.warn(
                "Parameter sample_points is deprecated. Use the "
                "parameter grid_points instead.",
                DeprecationWarning,
            )
            grid_points = sample_points

        grid_points = (
            self.grid_points
            if grid_points is None
            else _to_grid_points(grid_points)
        )

        return self.copy(
            data_matrix=self.evaluate(grid_points, grid=True),
            grid_points=grid_points,
        )

    def copy(  # noqa: WPS211
        self: T,
        *,
        deep: bool = False,  # For Pandas compatibility
        data_matrix: Optional[ArrayLike] = None,
        grid_points: Optional[GridPointsLike] = None,
        sample_points: Optional[GridPointsLike] = None,
        domain_range: Optional[DomainRangeLike] = None,
        dataset_name: Optional[str] = None,
        argument_names: Optional[LabelTupleLike] = None,
        coordinate_names: Optional[LabelTupleLike] = None,
        sample_names: Optional[LabelTupleLike] = None,
        extrapolation: Optional[ExtrapolationLike] = None,
        interpolation: Optional[Evaluator] = None,
    ) -> T:
        """
        Return a copy of the FDataGrid.

        If an argument is provided the corresponding attribute in the new copy
        is updated.

        """
        if sample_points is not None:
            warnings.warn(
                "Parameter sample_points is deprecated. Use the "
                "parameter grid_points instead.",
                DeprecationWarning,
            )
            grid_points = sample_points

        if data_matrix is None:
            # The data matrix won't be writeable
            data_matrix = self.data_matrix

        if grid_points is None:
            # Grid points won`t be writeable
            grid_points = self.grid_points

        if domain_range is None:
            domain_range = copy.deepcopy(self.domain_range)

        if dataset_name is None:
            dataset_name = self.dataset_name

        if argument_names is None:
            # Tuple, immutable
            argument_names = self.argument_names

        if coordinate_names is None:
            # Tuple, immutable
            coordinate_names = self.coordinate_names

        if sample_names is None:
            # Tuple, immutable
            sample_names = self.sample_names

        if extrapolation is None:
            extrapolation = self.extrapolation

        if interpolation is None:
            interpolation = self.interpolation

        return FDataGrid(
            data_matrix,
            grid_points=grid_points,
            domain_range=domain_range,
            dataset_name=dataset_name,
            argument_names=argument_names,
            coordinate_names=coordinate_names,
            sample_names=sample_names,
            extrapolation=extrapolation,
            interpolation=interpolation,
        )

    def restrict(
        self: T,
        domain_range: DomainRangeLike,
    ) -> T:
        """
        Restrict the functions to a new domain range.

        Args:
            domain_range: New domain range.

        Returns:
            Restricted function.

        """
        domain_range = _to_domain_range(domain_range)
        assert all(
            c <= a < b <= d  # noqa: WPS228
            for ((a, b), (c, d)) in zip(domain_range, self.domain_range)
        )

        index_list = []
        new_grid_points = []

        # Eliminate points outside the new range.
        for dr, grid_points in zip(
            domain_range,
            self.grid_points,
        ):
            keep_index = (
                (dr[0] <= grid_points)
                & (grid_points <= dr[1])
            )

            index_list.append(keep_index)

            new_grid_points.append(
                grid_points[keep_index],
            )

        data_matrix = self.data_matrix[tuple([slice(None)] + index_list)]

        return self.copy(
            domain_range=domain_range,
            grid_points=new_grid_points,
            data_matrix=data_matrix,
        )

    def shift(
        self,
        shifts: Union[ArrayLike, float],
        *,
        restrict_domain: bool = False,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid_points: Optional[GridPointsLike] = None,
    ) -> FDataGrid:
        r"""
        Perform a shift of the curves.

        The i-th shifted function :math:`y_i` has the form

        .. math::
            y_i(t) = x_i(t + \delta_i)

        where :math:`x_i` is the i-th original function and :math:`delta_i` is
        the shift performed for that function, that must be a vector in the
        domain space.

        Note that a positive shift moves the graph of the function in the
        negative direction and vice versa.

        Args:
            shifts: List with the shifts
                corresponding for each sample or numeric with the shift to
                apply to all samples.
            restrict_domain: If True restricts the domain to avoid the
                evaluation of points outside the domain using extrapolation.
                Defaults uses extrapolation.
            extrapolation: Controls the
                extrapolation mode for elements outside the domain range.
                By default uses the method defined in fd. See extrapolation to
                more information.
            grid_points: Grid of points where
                the functions are evaluated to obtain the discrete
                representation of the object to operate. If ``None`` the
                current grid_points are used to unificate the domain of the
                shifted data.

        Returns:
            Shifted functions.

        Examples:
            >>> import numpy as np
            >>> import skfda
            >>>
            >>> t = np.linspace(0, 1, 6)
            >>> x = np.array([t, t**2, t**3])
            >>> fd = FDataGrid(x, t)
            >>> fd.domain_range[0]
            (0.0, 1.0)
            >>> fd.grid_points[0]
            array([ 0. ,  0.2,  0.4,  0.6,  0.8,  1. ])
            >>> fd.data_matrix[..., 0]
            array([[ 0.   ,  0.2  ,  0.4  ,  0.6  ,  0.8  ,  1.   ],
                   [ 0.   ,  0.04 ,  0.16 ,  0.36 ,  0.64 ,  1.   ],
                   [ 0.   ,  0.008,  0.064,  0.216,  0.512,  1.   ]])

            Shift all curves by the same amount:

            >>> shifted = fd.shift(0.2)
            >>> shifted.domain_range[0]
            (0.0, 1.0)
            >>> shifted.grid_points[0]
            array([ 0. ,  0.2,  0.4,  0.6,  0.8,  1. ])
            >>> shifted.data_matrix[..., 0]
            array([[ 0.2  ,  0.4  ,  0.6  ,  0.8  ,  1.   ,  1.2  ],
                   [ 0.04 ,  0.16 ,  0.36 ,  0.64 ,  1.   ,  1.36 ],
                   [ 0.008,  0.064,  0.216,  0.512,  1.   ,  1.488]])


            Different shift per curve:

            >>> shifted = fd.shift([-0.2, 0.0, 0.2])
            >>> shifted.domain_range[0]
            (0.0, 1.0)
            >>> shifted.grid_points[0]
            array([ 0. ,  0.2,  0.4,  0.6,  0.8,  1. ])
            >>> shifted.data_matrix[..., 0]
            array([[-0.2  ,  0.   ,  0.2  ,  0.4  ,  0.6  ,  0.8  ],
                   [ 0.   ,  0.04 ,  0.16 ,  0.36 ,  0.64 ,  1.   ],
                   [ 0.008,  0.064,  0.216,  0.512,  1.   ,  1.488]])

            It is possible to restrict the domain to prevent the need for
            extrapolations:

            >>> shifted = fd.shift([-0.3, 0.1, 0.2], restrict_domain=True)
            >>> shifted.domain_range[0]
            (0.3, 0.8)

        """
        grid_points = (
            self.grid_points if grid_points is None
            else grid_points
        )

        return super().shift(
            shifts=shifts,
            restrict_domain=restrict_domain,
            extrapolation=extrapolation,
            grid_points=grid_points,
        )

    def compose(
        self: T,
        fd: T,
        *,
        eval_points: Optional[GridPointsLike] = None,
    ) -> T:
        """Composition of functions.

        Performs the composition of functions.

        Args:
            fd: FData object to make the composition. Should
                have the same number of samples and image dimension equal to 1.
            eval_points: Points to perform the evaluation.

        Returns:
            Function representing the composition.

        """
        if self.dim_domain != fd.dim_codomain:
            raise ValueError(
                f"Dimension of codomain of first function do not "
                f"match with the domain of the second function "
                f"{self.dim_domain} != {fd.dim_codomain}.",
            )

        # All composed with same function
        if fd.n_samples == 1 and self.n_samples != 1:
            fd = fd.copy(data_matrix=np.repeat(
                fd.data_matrix,
                self.n_samples,
                axis=0,
            ))

        if fd.dim_domain == 1:
            if eval_points is None:
                try:
                    eval_points = fd.grid_points[0]
                except AttributeError:
                    eval_points = np.linspace(
                        *fd.domain_range[0],
                        constants.N_POINTS_COARSE_MESH,
                    )

            eval_points_transformation = fd(eval_points)
            data_matrix = self(
                eval_points_transformation,
                aligned=False,
            )
        else:
            if eval_points is None:
                eval_points = fd.grid_points

            grid_transformation = fd(eval_points, grid=True)

            lengths = [len(ax) for ax in eval_points]

            eval_points_transformation = np.empty((
                self.n_samples,
                np.prod(lengths),
                self.dim_domain,
            ))

            for i in range(self.n_samples):
                eval_points_transformation[i] = np.array(
                    list(map(np.ravel, grid_transformation[i].T)),
                ).T

            data_matrix = self(
                eval_points_transformation,
                aligned=False,
            )

        return self.copy(
            data_matrix=data_matrix,
            grid_points=eval_points,
            domain_range=fd.domain_range,
            argument_names=fd.argument_names,
        )

    def __str__(self) -> str:
        """Return str(self)."""
        return (
            f"Data set:    {self.data_matrix}\n"
            f"grid_points:    {self.grid_points}\n"
            f"time range:    {self.domain_range}"
        )

    def __repr__(self) -> str:
        """Return repr(self)."""
        return (
            f"FDataGrid("  # noqa: WPS221
            f"\n{self.data_matrix!r},"
            f"\ngrid_points={self.grid_points!r},"
            f"\ndomain_range={self.domain_range!r},"
            f"\ndataset_name={self.dataset_name!r},"
            f"\nargument_names={self.argument_names!r},"
            f"\ncoordinate_names={self.coordinate_names!r},"
            f"\nextrapolation={self.extrapolation!r},"
            f"\ninterpolation={self.interpolation!r})"
        ).replace(
            '\n',
            '\n    ',
        )

    def __getitem__(self: T, key: Union[int, slice, np.ndarray]) -> T:
        """Return self[key]."""
        key = _check_array_key(self.data_matrix, key)

        return self.copy(
            data_matrix=self.data_matrix[key],
            sample_names=np.array(self.sample_names)[key],
        )

    #####################################################################
    # Numpy methods
    #####################################################################

    def __array_ufunc__(
        self,
        ufunc: Any,
        method: str,
        *inputs: Any,
        **kwargs: Any,
    ) -> Any:

        for i in inputs:
            if (
                isinstance(i, FDataGrid)
                and not np.array_equal(i.grid_points, self.grid_points)
            ):
                return NotImplemented

        new_inputs = [
            i.data_matrix if isinstance(i, FDataGrid)
            else self._get_op_matrix(i) for i in inputs
        ]

        outputs = kwargs.pop('out', None)
        if outputs:
            new_outputs = [
                o.data_matrix if isinstance(o, FDataGrid)
                else o for o in outputs
            ]
            kwargs['out'] = tuple(new_outputs)
        else:
            new_outputs = (None,) * ufunc.nout

        results = getattr(ufunc, method)(*new_inputs, **kwargs)
        if results is NotImplemented:
            return NotImplemented

        if ufunc.nout == 1:
            results = (results,)

        results = tuple(
            (result if output is None else output)
            for result, output in zip(results, new_outputs)
        )

        results = [self.copy(data_matrix=r) for r in results]

        return results[0] if len(results) == 1 else results

    #####################################################################
    # Pandas ExtensionArray methods
    #####################################################################

    def _take_allow_fill(
        self: T,
        indices: NDArrayInt,
        fill_value: T,
    ) -> T:
        result = self.copy()
        result.data_matrix = np.full(
            (len(indices),) + self.data_matrix.shape[1:],
            np.nan,
        )

        positive_mask = indices >= 0
        result.data_matrix[positive_mask] = self.data_matrix[
            indices[positive_mask]
        ]

        if fill_value is not self.dtype.na_value:
            result.data_matrix[~positive_mask] = fill_value.data_matrix[0]

        return result

    @property
    def dtype(self) -> FDataGridDType:
        """The dtype for this extension array, FDataGridDType"""
        return FDataGridDType(
            grid_points=self.grid_points,
            domain_range=self.domain_range,
            dim_codomain=self.dim_codomain,
        )

    @property
    def nbytes(self) -> int:
        """
        The number of bytes needed to store this object in memory.
        """
        return self.data_matrix.nbytes + sum(
            p.nbytes for p in self.grid_points
        )

    def isna(self) -> np.ndarray:
        """
        Return a 1-D array indicating if each value is missing.

        Returns:
            na_values: Positions of NA.
        """
        return np.all(
            np.isnan(self.data_matrix),
            axis=tuple(range(1, self.data_matrix.ndim)),
        )


class FDataGridDType(pandas.api.extensions.ExtensionDtype):  # type: ignore
    """DType corresponding to FDataGrid in Pandas."""

    name = 'FDataGrid'
    kind = 'O'
    type = FDataGrid  # noqa: WPS125
    na_value = pandas.NA

    def __init__(
        self,
        grid_points: GridPointsLike,
        dim_codomain: int,
        domain_range: Optional[DomainRangeLike] = None,
    ) -> None:
        grid_points = _to_grid_points(grid_points)

        self.grid_points = tuple(tuple(s) for s in grid_points)

        if domain_range is None:
            domain_range = tuple((s[0], s[-1]) for s in self.grid_points)

        self.domain_range = _to_domain_range(domain_range)
        self.dim_codomain = dim_codomain

    @classmethod
    def construct_array_type(cls) -> Type[FDataGrid]:  # noqa: D102
        return FDataGrid

    def _na_repr(self) -> FDataGrid:

        shape = (
            (1,)
            + tuple(len(s) for s in self.grid_points)
            + (self.dim_codomain,)
        )

        data_matrix = np.full(shape=shape, fill_value=np.NaN)

        return FDataGrid(
            grid_points=self.grid_points,
            domain_range=self.domain_range,
            data_matrix=data_matrix,
        )

    def __eq__(self, other: Any) -> bool:
        """
        Compare dtype equality.

        Rules for equality (similar to categorical):
        1) Any FData is equal to the string 'category'
        2) Any FData is equal to itself
        3) Otherwise, they are equal if the arguments are equal.
        6) Any other comparison returns False
        """
        if isinstance(other, str):
            return other == self.name
        elif other is self:
            return True

        return (
            isinstance(other, FDataGridDType)
            and self.dim_codomain == other.dim_codomain
            and self.domain_range == other.domain_range
            and self.grid_points == other.grid_points
        )

    def __hash__(self) -> int:
        return hash((self.grid_points, self.domain_range, self.dim_codomain))


class _CoordinateIterator(Sequence[T]):
    """Internal class to iterate through the image coordinates."""

    def __init__(self, fdatagrid: T) -> None:
        """Create an iterator through the image coordinates."""
        self._fdatagrid = fdatagrid

    def __getitem__(self, key: Union[int, slice]) -> T:
        """Get a specific coordinate."""
        s_key = key
        if isinstance(s_key, int):
            s_key = slice(s_key, s_key + 1)

        coordinate_names = np.array(self._fdatagrid.coordinate_names)[s_key]

        return self._fdatagrid.copy(
            data_matrix=self._fdatagrid.data_matrix[..., key],
            coordinate_names=coordinate_names,
        )

    def __len__(self) -> int:
        """Return the number of coordinates."""
        return self._fdatagrid.dim_codomain



"""Tests of basis functions."""

import itertools
import unittest

import numpy as np

import skfda
from skfda import concatenate
from skfda.misc import inner_product_matrix
from skfda.representation.basis import (
    BSpline,
    Constant,
    FDataBasis,
    Fourier,
    Monomial,
)
from skfda.representation.grid import FDataGrid


class TestBasis(unittest.TestCase):
    """Tests of basis and FDataBasis."""

    # def setUp(self): could be defined for set up before any test

    def test_from_data_cholesky(self) -> None:
        """Test basis conversion using Cholesky method."""
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = BSpline((0, 1), n_basis=5)
        np.testing.assert_array_almost_equal(
            FDataBasis.from_data(
                x,
                grid_points=t,
                basis=basis,
                method='cholesky',
            ).coefficients.round(2),
            np.array([[1.0, 2.78, -3.0, -0.78, 1.0]]),
        )

    def test_from_data_qr(self) -> None:
        """Test basis conversion using QR method."""
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = BSpline((0, 1), n_basis=5)
        np.testing.assert_array_almost_equal(
            FDataBasis.from_data(
                x,
                grid_points=t,
                basis=basis,
                method='qr',
            ).coefficients.round(2),
            np.array([[1.0, 2.78, -3.0, -0.78, 1.0]]),
        )

    def test_basis_inner_matrix(self) -> None:
        """Test the inner product matrix of FDataBasis objects."""
        basis = Monomial(n_basis=3)

        np.testing.assert_array_almost_equal(
            basis.inner_product_matrix(),
            [
                [1, 1 / 2, 1 / 3],  # noqa: WPS204
                [1 / 2, 1 / 3, 1 / 4],  # noqa: WPS204
                [1 / 3, 1 / 4, 1 / 5],
            ],
        )

        np.testing.assert_array_almost_equal(
            basis.inner_product_matrix(basis),
            [
                [1, 1 / 2, 1 / 3],
                [1 / 2, 1 / 3, 1 / 4],
                [1 / 3, 1 / 4, 1 / 5],
            ],
        )

        np.testing.assert_array_almost_equal(
            basis.inner_product_matrix(Monomial(n_basis=4)),
            [
                [1, 1 / 2, 1 / 3, 1 / 4],
                [1 / 2, 1 / 3, 1 / 4, 1 / 5],
                [1 / 3, 1 / 4, 1 / 5, 1 / 6],
            ],
        )

        # TODO testing with other basis

    def test_basis_gram_matrix_monomial(self) -> None:
        """Test the Gram matrix with monomial basis."""
        basis = Monomial(n_basis=3)
        gram_matrix = basis.gram_matrix()
        gram_matrix_numerical = basis._gram_matrix_numerical()  # noqa: WPS437
        gram_matrix_res = np.array([
            [1, 1 / 2, 1 / 3],
            [1 / 2, 1 / 3, 1 / 4],
            [1 / 3, 1 / 4, 1 / 5],
        ])

        np.testing.assert_allclose(
            gram_matrix,
            gram_matrix_res,
        )
        np.testing.assert_allclose(
            gram_matrix_numerical,
            gram_matrix_res,
        )

    def test_basis_gram_matrix_fourier(self) -> None:
        """Test the Gram matrix with fourier basis."""
        basis = Fourier(n_basis=3)
        gram_matrix = basis.gram_matrix()
        gram_matrix_numerical = basis._gram_matrix_numerical()  # noqa: WPS437
        gram_matrix_res = np.identity(3)

        np.testing.assert_allclose(
            gram_matrix,
            gram_matrix_res,
        )
        np.testing.assert_allclose(
            gram_matrix_numerical,
            gram_matrix_res,
            atol=1e-15,
            rtol=1e-15,
        )

    def test_basis_gram_matrix_bspline(self) -> None:
        """Test the Gram matrix with B-spline basis."""
        basis = BSpline(n_basis=6)
        gram_matrix = basis.gram_matrix()
        gram_matrix_numerical = basis._gram_matrix_numerical()  # noqa: WPS437
        gram_matrix_res = np.array([
            [0.04761905, 0.02916667, 0.00615079, 0.00039683, 0, 0],
            [0.02916667, 0.07380952, 0.05208333, 0.01145833, 0.00014881, 0],
            [  # noqa: WPS317
                0.00615079, 0.05208333, 0.10892857,
                0.07098214, 0.01145833, 0.00039683,
            ],
            [  # noqa: WPS317
                0.00039683, 0.01145833, 0.07098214,
                0.10892857, 0.05208333, 0.00615079,
            ],
            [0, 0.00014881, 0.01145833, 0.05208333, 0.07380952, 0.02916667],
            [0, 0, 0.00039683, 0.00615079, 0.02916667, 0.04761905],
        ])

        np.testing.assert_allclose(
            gram_matrix,
            gram_matrix_res,
            rtol=1e-4,
        )
        np.testing.assert_allclose(
            gram_matrix_numerical,
            gram_matrix_res,
            rtol=1e-4,
        )

    def test_basis_basis_inprod(self) -> None:
        """Test inner product between different basis."""
        monomial = Monomial(n_basis=4)
        bspline = BSpline(n_basis=5, order=4)
        np.testing.assert_allclose(
            monomial.inner_product_matrix(bspline),
            np.array([
                [0.12499983, 0.25000035, 0.24999965, 0.25000035, 0.12499983],
                [0.01249991, 0.07500017, 0.12499983, 0.17500017, 0.11249991],
                [0.00208338, 0.02916658, 0.07083342, 0.12916658, 0.10208338],
                [0.00044654, 0.01339264, 0.04375022, 0.09910693, 0.09330368],
            ]),
            rtol=1e-3,
        )
        np.testing.assert_array_almost_equal(
            monomial.inner_product_matrix(bspline),
            bspline.inner_product_matrix(monomial).T,
        )

    def test_basis_fdatabasis_inprod(self) -> None:
        """Test inner product between different basis expansions."""
        monomial = Monomial(n_basis=4)
        bspline = BSpline(n_basis=5, order=3)
        bsplinefd = FDataBasis(bspline, np.arange(0, 15).reshape(3, 5))

        np.testing.assert_allclose(
            inner_product_matrix(monomial.to_basis(), bsplinefd),
            np.array([
                [2.0, 7.0, 12.0],
                [1.29626206, 3.79626206, 6.29626206],
                [0.96292873, 2.62959539, 4.29626206],
                [0.7682873, 2.0182873, 3.2682873],
            ]),
            rtol=1e-4,
        )

    def test_fdatabasis_fdatabasis_inprod(self) -> None:
        """Test inner product between FDataBasis objects."""
        monomial = Monomial(n_basis=4)
        monomialfd = FDataBasis(
            monomial,
            [
                [5, 4, 1, 0],
                [4, 2, 1, 0],
                [4, 1, 6, 4],
                [4, 5, 0, 1],
                [5, 6, 2, 0],
            ],
        )
        bspline = BSpline(n_basis=5, order=3)
        bsplinefd = FDataBasis(bspline, np.arange(0, 15).reshape(3, 5))

        np.testing.assert_allclose(
            inner_product_matrix(monomialfd, bsplinefd),
            np.array([
                [16.14797697, 52.81464364, 89.4813103],
                [11.55565285, 38.22211951, 64.88878618],
                [18.14698361, 55.64698361, 93.14698361],
                [15.2495976, 48.9995976, 82.7495976],
                [19.70392982, 63.03676315, 106.37009648],
            ]),
            rtol=1e-4,
        )

    def test_comutativity_inprod(self) -> None:
        """Test commutativity of the inner product."""
        monomial = Monomial(n_basis=4)
        bspline = BSpline(n_basis=5, order=3)
        bsplinefd = FDataBasis(bspline, np.arange(0, 15).reshape(3, 5))

        np.testing.assert_allclose(
            inner_product_matrix(bsplinefd, monomial.to_basis()),
            np.transpose(inner_product_matrix(monomial.to_basis(), bsplinefd)),
        )

    def test_concatenate(self) -> None:
        """Test concatenation of two FDataBasis."""
        sample1 = np.arange(0, 10)
        sample2 = np.arange(10, 20)
        fd1 = FDataGrid([sample1]).to_basis(Fourier(n_basis=5))
        fd2 = FDataGrid([sample2]).to_basis(Fourier(n_basis=5))

        fd = concatenate([fd1, fd2])

        np.testing.assert_equal(fd.n_samples, 2)
        np.testing.assert_equal(fd.dim_codomain, 1)
        np.testing.assert_equal(fd.dim_domain, 1)
        np.testing.assert_array_equal(
            fd.coefficients,
            np.concatenate([fd1.coefficients, fd2.coefficients]),
        )


class TestFDataBasisOperations(unittest.TestCase):
    """Test FDataBasis operations."""

    def test_fdatabasis_add(self) -> None:
        """Test addition of FDataBasis."""
        monomial1 = FDataBasis(Monomial(n_basis=3), [1, 2, 3])
        monomial2 = FDataBasis(Monomial(n_basis=3), [[1, 2, 3], [3, 4, 5]])

        self.assertTrue(
            (monomial1 + monomial2).equals(
                FDataBasis(
                    Monomial(n_basis=3),
                    [[2, 4, 6], [4, 6, 8]],
                ),
            ),
        )

        with np.testing.assert_raises(TypeError):
            monomial2 + FDataBasis(  # noqa: WPS428
                Fourier(n_basis=3),
                [[2, 2, 3], [5, 4, 5]],
            )

    def test_fdatabasis_sub(self) -> None:
        """Test subtraction of FDataBasis."""
        monomial1 = FDataBasis(Monomial(n_basis=3), [1, 2, 3])
        monomial2 = FDataBasis(Monomial(n_basis=3), [[1, 2, 3], [3, 4, 5]])

        self.assertTrue(
            (monomial1 - monomial2).equals(
                FDataBasis(
                    Monomial(n_basis=3),
                    [[0, 0, 0], [-2, -2, -2]],
                ),
            ),
        )

        with np.testing.assert_raises(TypeError):
            monomial2 - FDataBasis(  # noqa: WPS428
                Fourier(n_basis=3),
                [[2, 2, 3], [5, 4, 5]],
            )

    def test_fdatabasis_mul(self) -> None:
        """Test multiplication of FDataBasis."""
        basis = Monomial(n_basis=3)

        monomial1 = FDataBasis(basis, [1, 2, 3])
        monomial2 = FDataBasis(basis, [[1, 2, 3], [3, 4, 5]])

        self.assertTrue(
            (monomial1 * 2).equals(
                FDataBasis(
                    basis,
                    [[2, 4, 6]],
                ),
            ),
        )

        self.assertTrue(
            (3 * monomial2).equals(
                FDataBasis(
                    basis,
                    [[3, 6, 9], [9, 12, 15]],
                ),
            ),
        )

        self.assertTrue(
            (3 * monomial2).equals(
                monomial2 * 3,
            ),
        )

        self.assertTrue(
            (monomial2 * [1, 2]).equals(
                FDataBasis(
                    basis,
                    [[1, 2, 3], [6, 8, 10]],
                ),
            ),
        )

        self.assertTrue(
            ([1, 2] * monomial2).equals(
                FDataBasis(
                    basis,
                    [[1, 2, 3], [6, 8, 10]],
                ),
            ),
        )

        with np.testing.assert_raises(TypeError):
            monomial2 * FDataBasis(  # noqa: WPS428
                Fourier(n_basis=3),
                [[2, 2, 3], [5, 4, 5]],
            )

        with np.testing.assert_raises(TypeError):
            monomial2 * monomial2  # noqa: WPS428

    def test_fdatabasis_div(self) -> None:
        """Test division of FDataBasis."""
        basis = Monomial(n_basis=3)

        monomial1 = FDataBasis(basis, [1, 2, 3])
        monomial2 = FDataBasis(basis, [[1, 2, 3], [3, 4, 5]])

        self.assertTrue((monomial1 / 2).equals(
            FDataBasis(
                basis,
                [[1 / 2, 1, 3 / 2]],
            ),
        ))

        self.assertTrue(
            (monomial2 / 2).equals(
                FDataBasis(
                    basis,
                    [[1 / 2, 1, 3 / 2], [3 / 2, 2, 5 / 2]],
                ),
            ),
        )

        self.assertTrue(
            (monomial2 / [1, 2]).equals(
                FDataBasis(
                    basis,
                    [[1.0, 2.0, 3.0], [3 / 2, 2, 5 / 2]],
                ),
            ),
        )


class TestFDataBasisDerivatives(unittest.TestCase):
    """Test FDataBasis derivatives."""

    def test_fdatabasis_derivative_constant(self) -> None:
        """Test derivatives with a constant basis."""
        constant = FDataBasis(
            Constant(),
            [[1], [2], [3], [4]],
        )

        self.assertTrue(
            constant.derivative().equals(
                FDataBasis(
                    Constant(),
                    [[0], [0], [0], [0]],
                ),
            ),
        )

        self.assertTrue(
            constant.derivative(order=0).equals(
                FDataBasis(
                    Constant(),
                    [[1], [2], [3], [4]],
                ),
            ),
        )

    def test_fdatabasis_derivative_monomial(self) -> None:
        """Test derivatives with a monomial basis."""
        monomial = FDataBasis(
            Monomial(n_basis=8),
            [1, 5, 8, 9, 7, 8, 4, 5],
        )

        monomial2 = FDataBasis(
            Monomial(n_basis=5),
            [
                [4, 9, 7, 4, 3],
                [1, 7, 9, 8, 5],
                [4, 6, 6, 6, 8],
            ],
        )

        self.assertTrue(
            monomial.derivative().equals(
                FDataBasis(
                    Monomial(n_basis=7),
                    [5, 16, 27, 28, 40, 24, 35],
                ),
            ),
        )

        self.assertTrue(
            monomial.derivative(order=0).equals(monomial),
        )

        self.assertTrue(
            monomial.derivative(order=6).equals(
                FDataBasis(
                    Monomial(n_basis=2),
                    [2880, 25200],
                ),
            ),
        )

        self.assertTrue(
            monomial2.derivative().equals(
                FDataBasis(
                    Monomial(n_basis=4),
                    [
                        [9, 14, 12, 12],
                        [7, 18, 24, 20],
                        [6, 12, 18, 32],
                    ],
                ),
            ),
        )

        self.assertTrue(
            monomial2.derivative(order=0).equals(monomial2),
        )

        self.assertTrue(
            monomial2.derivative(order=3).equals(
                FDataBasis(
                    Monomial(n_basis=2),
                    [
                        [24, 72],
                        [48, 120],
                        [36, 192],
                    ],
                ),
            ),
        )

    def test_fdatabasis_derivative_fourier(self) -> None:
        """Test derivatives with a fourier basis."""
        fourier = FDataBasis(
            Fourier(n_basis=7),
            [1, 5, 8, 9, 8, 4, 5],
        )

        fourier2 = FDataBasis(
            Fourier(n_basis=5),
            [
                [4, 9, 7, 4, 3],
                [1, 7, 9, 8, 5],
                [4, 6, 6, 6, 8],
            ],
        )

        fou0 = fourier.derivative(order=0)
        fou1 = fourier.derivative()
        fou2 = fourier.derivative(order=2)

        np.testing.assert_equal(fou1.basis, fourier.basis)
        np.testing.assert_almost_equal(
            fou1.coefficients.round(5),
            np.atleast_2d(
                [  # noqa: WPS317
                    0, -50.26548, 31.41593, -100.53096,
                    113.09734, -94.24778, 75.39822,
                ],
            ),
        )

        self.assertTrue(fou0.equals(fourier))
        np.testing.assert_equal(fou2.basis, fourier.basis)
        np.testing.assert_almost_equal(
            fou2.coefficients.round(5),
            np.atleast_2d(
                [  # noqa: WPS317
                    0, -197.39209, -315.82734, -1421.22303,
                    -1263.30936, -1421.22303, -1776.52879,
                ],
            ),
        )

        fou0 = fourier2.derivative(order=0)
        fou1 = fourier2.derivative()
        fou2 = fourier2.derivative(order=2)

        np.testing.assert_equal(fou1.basis, fourier2.basis)
        np.testing.assert_almost_equal(
            fou1.coefficients.round(5),
            [
                [0, -43.9823, 56.54867, -37.69911, 50.26548],
                [0, -56.54867, 43.9823, -62.83185, 100.53096],
                [0, -37.69911, 37.69911, -100.53096, 75.39822],
            ],
        )

        self.assertTrue(fou0.equals(fourier2))
        np.testing.assert_equal(fou2.basis, fourier2.basis)
        np.testing.assert_almost_equal(
            fou2.coefficients.round(5),
            [
                [0, -355.30576, -276.34892, -631.65468, -473.74101],
                [0, -276.34892, -355.30576, -1263.30936, -789.56835],
                [0, -236.87051, -236.87051, -947.48202, -1263.30936],
            ],
        )

    def test_fdatabasis_derivative_bspline(self) -> None:
        """Test derivatives with a B-spline basis."""
        bspline = FDataBasis(
            BSpline(n_basis=8),
            [1, 5, 8, 9, 7, 8, 4, 5],
        )
        bspline2 = FDataBasis(
            BSpline(n_basis=5),
            [
                [4, 9, 7, 4, 3],
                [1, 7, 9, 8, 5],
                [4, 6, 6, 6, 8],
            ],
        )

        bs0 = bspline.derivative(order=0)
        bs1 = bspline.derivative()
        bs2 = bspline.derivative(order=2)
        np.testing.assert_equal(bs1.basis, BSpline(n_basis=7, order=3))

        np.testing.assert_almost_equal(
            bs1.coefficients,
            np.atleast_2d([60, 22.5, 5, -10, 5, -30, 15]),
        )

        self.assertTrue(bs0.equals(bspline))

        np.testing.assert_equal(
            bs2.basis,
            BSpline(n_basis=6, order=2),
        )

        np.testing.assert_almost_equal(
            bs2.coefficients,
            np.atleast_2d([-375, -87.5, -75, 75, -175, 450]),
        )

        bs0 = bspline2.derivative(order=0)
        bs1 = bspline2.derivative()
        bs2 = bspline2.derivative(order=2)

        np.testing.assert_equal(bs1.basis, BSpline(n_basis=4, order=3))

        np.testing.assert_almost_equal(
            bs1.coefficients,
            [
                [30, -6, -9, -6],
                [36, 6, -3, -18],
                [12, 0, 0, 12],
            ],
        )

        self.assertTrue(bs0.equals(bspline2))

        np.testing.assert_equal(
            bs2.basis,
            BSpline(n_basis=3, order=2),
        )

        np.testing.assert_almost_equal(
            bs2.coefficients,
            [
                [-144, -6, 12],
                [-120, -18, -60],
                [-48, 0, 48],
            ],
        )


class TestVectorValuedBasis(unittest.TestCase):
    """Tests for the vector valued basis."""

    def test_vector_valued(self) -> None:
        """Test vector valued basis."""
        X, _ = skfda.datasets.fetch_weather(return_X_y=True)

        basis_dim = skfda.representation.basis.Fourier(
            n_basis=7,
            domain_range=X.domain_range,
        )
        basis = skfda.representation.basis.VectorValued(
            [basis_dim] * 2,
        )

        X_basis = X.to_basis(basis)

        self.assertEqual(X_basis.dim_codomain, 2)

        self.assertEqual(X_basis.coordinates[0].basis, basis_dim)
        np.testing.assert_allclose(
            X_basis.coordinates[0].coefficients,
            X.coordinates[0].to_basis(basis_dim).coefficients,
        )

        self.assertEqual(X_basis.coordinates[1].basis, basis_dim)
        np.testing.assert_allclose(
            X_basis.coordinates[1].coefficients,
            X.coordinates[1].to_basis(basis_dim).coefficients,
        )


class TestTensorBasis(unittest.TestCase):
    """Tests for the Tensor basis."""

    def setUp(self) -> None:
        """Create original and tensor bases."""
        self.n_x = 4
        self.n_y = 3
        self.n_z = 5

        self.n = self.n_x * self.n_y * self.n_z

        self.dims = (self.n_x, self.n_y, self.n_z)

        self.basis_x = skfda.representation.basis.Monomial(n_basis=self.n_x)
        self.basis_y = skfda.representation.basis.Fourier(n_basis=self.n_y)
        self.basis_z = skfda.representation.basis.BSpline(n_basis=self.n_z)

        self.basis = skfda.representation.basis.Tensor([
            self.basis_x,
            self.basis_y,
            self.basis_z,
        ])

    def test_tensor_order(self) -> None:
        """
        Check the order of the elements in the tensor basis.

        The order should be:

        a_1 b_1 c_1, a_1 b_1 c_2, ..., a_1 b_1 c_n,
        a_1 b_2 c_1, a_1 b_1 c_2, ..., a_1 b_2 c_n,
        .
        .
        .
        a_1 b_m c_1, a_1 b_1 c_2, ..., a_1 b_m c_n,
        a_2 b_1 c_1, a_2 b_1 c_2, ..., a_2 b_1 c_n,
        .
        .
        .

        where the bases of the original spaces are A, B and C.

        """
        x_vals = [0, 0.3, 0.7]
        y_vals = [0.2, 0.5, 0.9]
        z_vals = [0.1, 0.4, 0.8]

        for t in itertools.product(x_vals, y_vals, z_vals):

            val_x = self.basis_x(t[0])
            val_y = self.basis_y(t[1])
            val_z = self.basis_z(t[2])
            val = self.basis([t])

            for x, y, z in itertools.product(
                range(self.n_x),
                range(self.n_y),
                range(self.n_z),
            ):

                index = (
                    x * self.n_y * self.n_z
                    + y * self.n_z
                    + z
                )

                index2 = np.ravel_multi_index(
                    [x, y, z],
                    dims=self.dims,
                )

                self.assertEqual(index, index2)

                self.assertAlmostEqual(
                    val[index],
                    val_x[x] * val_y[y] * val_z[z],
                )

    def test_tensor_gram_matrix(self) -> None:
        """Check that the Gram matrix is right."""
        gram_x = self.basis_x.gram_matrix()
        gram_y = self.basis_y.gram_matrix()
        gram_z = self.basis_z.gram_matrix()

        gram = self.basis.gram_matrix()

        for i in range(self.n):
            for j in range(self.n):
                left = np.unravel_index(i, shape=self.dims)
                right = np.unravel_index(j, shape=self.dims)

                value_gram = gram[i, j]
                value_gram_x = gram_x[left[0], right[0]]
                value_gram_y = gram_y[left[1], right[1]]
                value_gram_z = gram_z[left[2], right[2]]

                self.assertAlmostEqual(
                    value_gram,
                    value_gram_x * value_gram_y * value_gram_z,
                )


if __name__ == '__main__':
    unittest.main()



"""Tests of classification methods."""

import unittest

import numpy as np
from sklearn.base import clone
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsClassifier as _KNeighborsClassifier

from skfda.datasets import fetch_growth
from skfda.misc.metrics import l2_distance
from skfda.ml.classification import (
    DDClassifier,
    DDGClassifier,
    DTMClassifier,
    KNeighborsClassifier,
    MaximumDepthClassifier,
    NearestCentroid,
    RadiusNeighborsClassifier,
)
from skfda.ml.classification._depth_classifiers import _ArgMaxClassifier
from skfda.representation import FData


class TestClassifiers(unittest.TestCase):
    """Tests for classifiers."""

    def setUp(self) -> None:
        """Establish train and test data sets."""
        X, y = fetch_growth(return_X_y=True)
        X_train, X_test, y_train, y_test = train_test_split(
            X,
            y,
            test_size=0.25,
            stratify=y,
            random_state=0,
        )
        self._X_train = X_train
        self._X_test = X_test
        self._y_train = y_train
        self._y_test = y_test

    def test_dtm_independent_copy(self) -> None:
        """Check that copies are un-linked."""
        clf: DTMClassifier[FData] = DTMClassifier(proportiontocut=0.25)
        clf1 = clone(clf)
        clf2: DTMClassifier[FData] = DTMClassifier(proportiontocut=0.75)

        clf1.proportiontocut = 0.75
        clf1.fit(self._X_train, self._y_train)
        clf2.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf1.predict(self._X_test),
            clf2.predict(self._X_test),
        )

    def test_dtm_classifier(self) -> None:
        """Check DTM classifier."""
        clf: DTMClassifier[FData] = DTMClassifier(proportiontocut=0.25)
        clf.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf.predict(self._X_test),
            [  # noqa: WPS317
                1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1,
            ],
        )

    def test_centroid_classifier(self) -> None:
        """Check NearestCentroid classifier."""
        clf: NearestCentroid[FData] = NearestCentroid()
        clf.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf.predict(self._X_test),
            [  # noqa: WPS317
                1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
                1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1,
            ],
        )

    def test_dtm_inheritance(self) -> None:
        """Check that DTM is a subclass of NearestCentroid."""
        clf1: NearestCentroid[FData] = NearestCentroid()
        clf2: DTMClassifier[FData] = DTMClassifier(
            proportiontocut=0,
            metric=l2_distance,
        )
        clf1.fit(self._X_train, self._y_train)
        clf2.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf1.predict(self._X_test),
            clf2.predict(self._X_test),
        )

    def test_maximumdepth_classifier(self) -> None:
        """Check MaximumDepth classifier."""
        clf: MaximumDepthClassifier[FData] = MaximumDepthClassifier()
        clf.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf.predict(self._X_test),
            [  # noqa: WPS317
                1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1,
            ],
        )

    def test_dd_classifier(self) -> None:
        """Check DD classifier."""
        clf: DDClassifier[FData] = DDClassifier(degree=2)
        clf.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf.predict(self._X_test),
            [  # noqa: WPS317
                1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1,
            ],
        )

    def test_ddg_classifier(self) -> None:
        """Check DDG classifier."""
        clf: DDGClassifier[FData] = DDGClassifier(_KNeighborsClassifier())
        clf.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf.predict(self._X_test),
            [  # noqa: WPS317
                1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1,
                1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1,
            ],
        )

    def test_maximumdepth_inheritance(self) -> None:
        """Check that MaximumDepth is a subclass of DDG."""
        clf1: DDGClassifier[FData] = DDGClassifier(_ArgMaxClassifier())
        clf2: MaximumDepthClassifier[FData] = MaximumDepthClassifier()
        clf1.fit(self._X_train, self._y_train)
        clf2.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf1.predict(self._X_test),
            clf2.predict(self._X_test),
        )

    def test_kneighbors_classifier(self) -> None:
        """Check KNeighbors classifier."""
        clf = KNeighborsClassifier()
        clf.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf.predict(self._X_test),
            [  # noqa: WPS317
                0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1,
            ],
        )

    def test_radiusneighbors_classifier(self) -> None:
        """Check RadiusNeighbors classifier."""
        clf = RadiusNeighborsClassifier(radius=15)
        clf.fit(self._X_train, self._y_train)

        np.testing.assert_array_equal(
            clf.predict(self._X_test),
            [  # noqa: WPS317
                0, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 1,
                1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1,
            ],
        )

    def test_radiusneighbors_small_raidus(self) -> None:
        """Check that an error is raised if radius too small."""
        clf = RadiusNeighborsClassifier(radius=1)
        clf.fit(self._X_train, self._y_train)

        with np.testing.assert_raises(ValueError):
            clf.predict(self._X_test)


if __name__ == '__main__':
    unittest.main()



"""Registration of functional data module.

This module contains routines related to the registration procedure.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import numpy as np
from scipy.interpolate import PchipInterpolator

from ..representation._typing import ArrayLike, DomainRangeLike, NDArrayFloat
from ._utils import _to_domain_range, check_is_univariate

if TYPE_CHECKING:
    from ..representation import FDataGrid


def invert_warping(
    warping: FDataGrid,
    *,
    output_points: Optional[ArrayLike] = None,
) -> FDataGrid:
    r"""
    Compute the inverse of a diffeomorphism.

    Let :math:`\gamma : [a,b] \rightarrow [a,b]` be a function strictly
    increasing, calculates the corresponding inverse
    :math:`\gamma^{-1} : [a,b] \rightarrow [a,b]` such that
    :math:`\gamma^{-1} \circ \gamma = \gamma \circ \gamma^{-1} = \gamma_{id}`.

    Uses a PCHIP interpolator to compute approximately the inverse.

    Args:
        warping: Functions to be inverted.
        output_points: Set of points where the
            functions are interpolated to obtain the inverse, by default uses
            the sample points of the fdatagrid.

    Returns:
        Inverse of the original functions.

    Raises:
        ValueError: If the functions are not strictly increasing or are
            multidimensional.

    Examples:
        >>> import numpy as np
        >>> from skfda import FDataGrid

        We will construct the warping :math:`\gamma : [0,1] \rightarrow [0,1]`
        wich maps t to t^3.

        >>> t = np.linspace(0, 1)
        >>> gamma = FDataGrid(t**3, t)
        >>> gamma
        FDataGrid(...)

        We will compute the inverse.

        >>> inverse = invert_warping(gamma)
        >>> inverse
        FDataGrid(...)

        The result of the composition should be approximately the identity
        function .

        >>> identity = gamma.compose(inverse)
        >>> identity([0, 0.25, 0.5, 0.75, 1]).round(3)
        array([[[ 0.  ],
                [ 0.25],
                [ 0.5 ],
                [ 0.75],
                [ 1.  ]]])

    """
    check_is_univariate(warping)

    output_points = (
        warping.grid_points[0]
        if output_points is None
        else np.asarray(output_points)
    )

    y = warping(output_points)[..., 0]

    data_matrix = np.empty((warping.n_samples, len(output_points)))

    for i in range(warping.n_samples):
        data_matrix[i] = PchipInterpolator(y[i], output_points)(output_points)

    return warping.copy(data_matrix=data_matrix, grid_points=output_points)


def normalize_scale(
    t: NDArrayFloat,
    a: float = 0,
    b: float = 1,
) -> NDArrayFloat:
    """
    Perfoms an afine translation to normalize an interval.

    Args:
        t: Array of dim 1 or 2 with at least 2 values.
        a: Starting point of the new interval. Defaults 0.
        b: Stopping point of the new interval. Defaults 1.

    Returns:
        Array with the transformed interval.

    """
    t = t.T  # Broadcast to normalize multiple arrays
    t1 = (t - t[0]).astype(float)  # Translation to [0, t[-1] - t[0]]
    t1 *= (b - a) / (t[-1] - t[0])  # Scale to [0, b-a]
    t1 += a  # Translation to [a, b]
    t1[0] = a  # Fix possible round errors
    t1[-1] = b

    return t1.T


def normalize_warping(
    warping: FDataGrid,
    domain_range: Optional[DomainRangeLike] = None,
) -> FDataGrid:
    r"""
    Rescale a warping to normalize their :term:`domain`.

    Given a set of warpings :math:`\gamma_i:[a,b]\rightarrow  [a,b]` it is
    used an affine traslation to change the domain of the transformation to
    other domain, :math:`\tilde \gamma_i:[\tilde a,\tilde b] \rightarrow
    [\tilde a, \tilde b]`.

    Args:
        warping: Set of warpings to rescale.
        domain_range: New domain range of the warping. By
            default it is used the same domain range.

    Returns:
        Normalized warpings.

    """
    domain_range_tuple = (
        warping.domain_range[0]
        if domain_range is None
        else _to_domain_range(domain_range)[0]
    )

    data_matrix = normalize_scale(
        warping.data_matrix[..., 0],
        *domain_range_tuple,
    )
    grid_points = normalize_scale(warping.grid_points[0], *domain_range_tuple)

    return warping.copy(
        data_matrix=data_matrix,
        grid_points=grid_points,
        domain_range=domain_range,
    )



"""Test regularization methods."""
from __future__ import annotations

import unittest
import warnings
from typing import Callable, Optional, Sequence, Union

import numpy as np
from sklearn.datasets import make_regression
from sklearn.linear_model import Ridge
from sklearn.model_selection._split import train_test_split

import skfda
from skfda.misc.operators import (
    Identity,
    LinearDifferentialOperator,
    gramian_matrix,
)
from skfda.misc.operators._linear_differential_operator import (
    _monomial_evaluate_constant_linear_diff_op,
)
from skfda.misc.operators._operators import gramian_matrix_numerical
from skfda.misc.regularization import L2Regularization
from skfda.ml.regression import LinearRegression
from skfda.representation.basis import (
    Basis,
    BSpline,
    Constant,
    Fourier,
    Monomial,
)

LinearDifferentialOperatorInput = Union[
    int,
    Sequence[Union[float, Callable[[np.ndarray], np.ndarray]]],
    None,
]


class TestLinearDifferentialOperatorRegularization(unittest.TestCase):
    """Test linear differential operator penalty with different bases."""

    def _test_penalty(
        self,
        basis: Basis,
        linear_diff_op: LinearDifferentialOperatorInput,
        atol: float = 0,
        result: Optional[np.ndarray] = None,
    ) -> None:

        operator = LinearDifferentialOperator(linear_diff_op)

        penalty = gramian_matrix(operator, basis)
        numerical_penalty = gramian_matrix_numerical(operator, basis)

        np.testing.assert_allclose(
            penalty,
            numerical_penalty,
            atol=atol,
        )

        if result is not None:
            np.testing.assert_allclose(
                penalty,
                result,
                atol=atol,
            )

    def test_constant_penalty(self) -> None:
        """Test penalty for Constant basis."""
        basis = Constant(domain_range=(0, 3))

        res = np.array([[12]])

        self._test_penalty(basis, linear_diff_op=[2, 3, 4], result=res)

    def test_monomial_linear_diff_op(self) -> None:
        """Test directly the penalty for Monomial basis."""
        n_basis = 5

        basis = Monomial(n_basis=n_basis)

        linear_diff_op = [3]
        res = np.array([
            [0, 0, 0, 0, 3],
            [0, 0, 0, 3, 0],
            [0, 0, 3, 0, 0],
            [0, 3, 0, 0, 0],
            [3, 0, 0, 0, 0],
        ])

        np.testing.assert_allclose(
            _monomial_evaluate_constant_linear_diff_op(
                basis,
                np.array(linear_diff_op),
            ),
            res,
        )

        linear_diff_op = [3, 2]
        res = np.array([
            [0, 0, 0, 0, 3],
            [0, 0, 0, 3, 2],
            [0, 0, 3, 4, 0],
            [0, 3, 6, 0, 0],
            [3, 8, 0, 0, 0],
        ])

        np.testing.assert_allclose(
            _monomial_evaluate_constant_linear_diff_op(
                basis,
                np.array(linear_diff_op),
            ),
            res,
        )

        linear_diff_op = [3, 0, 5]
        res = np.array([
            [0, 0, 0, 0, 3],
            [0, 0, 0, 3, 0],
            [0, 0, 3, 0, 10],
            [0, 3, 0, 30, 0],
            [3, 0, 60, 0, 0],
        ])

        np.testing.assert_allclose(
            _monomial_evaluate_constant_linear_diff_op(
                basis,
                np.array(linear_diff_op),
            ),
            res,
        )

    def test_monomial_penalty(self) -> None:
        """Test penalty for Monomial basis."""
        basis = Monomial(n_basis=5, domain_range=(0, 3))

        # Theorethical result
        res = np.array([
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 12, 54, 216],
            [0, 0, 54, 324, 1458],
            [0, 0, 216, 1458, 6998.4],
        ])

        self._test_penalty(basis, linear_diff_op=2, result=res)

        basis = Monomial(n_basis=8, domain_range=(1, 5))

        self._test_penalty(basis, linear_diff_op=[1, 2, 3])
        self._test_penalty(basis, linear_diff_op=7)
        self._test_penalty(basis, linear_diff_op=0)
        self._test_penalty(basis, linear_diff_op=1)
        self._test_penalty(basis, linear_diff_op=27)

    def test_fourier_penalty(self) -> None:
        """Test penalty for Fourier basis."""
        basis = Fourier(n_basis=5)

        res = np.array([
            [0, 0, 0, 0, 0],
            [0, 1558.55, 0, 0, 0],
            [0, 0, 1558.55, 0, 0],
            [0, 0, 0, 24936.73, 0],
            [0, 0, 0, 0, 24936.73],
        ])

        # Those comparisons require atol as there are zeros involved
        self._test_penalty(basis, linear_diff_op=2, atol=0.01, result=res)

        basis = Fourier(n_basis=9, domain_range=(1, 5))
        self._test_penalty(basis, linear_diff_op=[1, 2, 3], atol=1e-7)
        self._test_penalty(basis, linear_diff_op=[2, 3, 0.1, 1], atol=1e-7)
        self._test_penalty(basis, linear_diff_op=0, atol=1e-7)
        self._test_penalty(basis, linear_diff_op=1, atol=1e-7)
        self._test_penalty(basis, linear_diff_op=3, atol=1e-7)

    def test_bspline_penalty(self) -> None:
        """Test penalty for BSpline basis."""
        basis = BSpline(n_basis=5)

        res = np.array([
            [96, -132, 24, 12, 0],
            [-132, 192, -48, -24, 12],
            [24, -48, 48, -48, 24],
            [12, -24, -48, 192, -132],
            [0, 12, 24, -132, 96],
        ])

        self._test_penalty(basis, linear_diff_op=2, result=res)

        basis = BSpline(n_basis=9, domain_range=(1, 5))
        self._test_penalty(basis, linear_diff_op=[1, 2, 3])
        self._test_penalty(basis, linear_diff_op=[2, 3, 0.1, 1])
        self._test_penalty(basis, linear_diff_op=0)
        self._test_penalty(basis, linear_diff_op=1)
        self._test_penalty(basis, linear_diff_op=3)
        self._test_penalty(basis, linear_diff_op=4)

        basis = BSpline(n_basis=16, order=8)
        self._test_penalty(basis, linear_diff_op=0, atol=1e-7)

    def test_bspline_penalty_special_case(self) -> None:
        """Test for behavior like in issue #185."""
        basis = BSpline(n_basis=5)

        res = np.array([
            [1152, -2016, 1152, -288, 0],
            [-2016, 3600, -2304, 1008, -288],
            [1152, -2304, 2304, -2304, 1152],
            [-288, 1008, -2304, 3600, -2016],
            [0, -288, 1152, -2016, 1152],
        ])

        operator = LinearDifferentialOperator(basis.order - 1)
        penalty = gramian_matrix(operator, basis)
        numerical_penalty = gramian_matrix_numerical(operator, basis)

        np.testing.assert_allclose(
            penalty,
            res,
        )

        np.testing.assert_allclose(
            numerical_penalty,
            res,
        )


class TestDefaultL2Regularization(unittest.TestCase):
    """Test default value of L2 regularization."""

    def test_basis_default(self) -> None:
        """Test that in basis smoothing."""
        data_matrix = np.linspace([0, 1, 2, 3], [1, 2, 3, 4], 100)

        fd = skfda.FDataGrid(data_matrix.T)

        smoother = skfda.preprocessing.smoothing.BasisSmoother(
            basis=skfda.representation.basis.BSpline(
                n_basis=10,
                domain_range=fd.domain_range,
            ),
            regularization=L2Regularization(),
        )

        smoother2 = skfda.preprocessing.smoothing.BasisSmoother(
            basis=skfda.representation.basis.BSpline(
                n_basis=10,
                domain_range=fd.domain_range,
            ),
            regularization=L2Regularization(
                Identity(),
            ),
        )

        fd_basis = smoother.fit_transform(fd)
        fd_basis2 = smoother2.fit_transform(fd)

        np.testing.assert_allclose(
            fd_basis.data_matrix,
            fd_basis2.data_matrix,
        )


class TestEndpointsDifferenceRegularization(unittest.TestCase):
    """Test regularization with a callable."""

    def test_basis_conversion(self) -> None:
        """Test that in basis smoothing."""
        data_matrix = np.linspace([0, 1, 2, 3], [1, 2, 3, 4], 100)

        fd = skfda.FDataGrid(data_matrix.T)

        smoother = skfda.preprocessing.smoothing.BasisSmoother(
            basis=skfda.representation.basis.BSpline(
                n_basis=10,
                domain_range=fd.domain_range,
            ),
            regularization=L2Regularization(
                lambda x: x(1)[:, 0] - x(0)[:, 0],
            ),
            smoothing_parameter=10000,
        )

        fd_basis = smoother.fit_transform(fd)

        np.testing.assert_allclose(
            fd_basis(0),
            fd_basis(1),
            atol=0.001,
        )


class TestL2Regularization(unittest.TestCase):
    """Test the L2 regularization."""

    def test_multivariate(self) -> None:
        """Test that it works with multivariate inputs."""

        def ignore_scalar_warning() -> None:  # noqa: WPS430
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="All the covariates are scalar.",
            )

        X, y = make_regression(
            n_samples=20,
            n_features=10,
            random_state=1,
            bias=3.5,
        )

        X_train, X_test, y_train, _ = train_test_split(
            X,
            y,
            random_state=2,
        )

        for regularization_parameter in (0, 1, 10, 100):

            with self.subTest(
                regularization_parameter=regularization_parameter,
            ):

                sklearn_l2 = Ridge(alpha=regularization_parameter)
                skfda_l2 = LinearRegression(
                    regularization=L2Regularization(
                        regularization_parameter=regularization_parameter,
                    ),
                )

                sklearn_l2.fit(X_train, y_train)
                with warnings.catch_warnings():
                    ignore_scalar_warning()
                    skfda_l2.fit(X_train, y_train)

                sklearn_y_pred = sklearn_l2.predict(X_test)
                with warnings.catch_warnings():
                    ignore_scalar_warning()
                    skfda_y_pred = skfda_l2.predict(X_test)

                np.testing.assert_allclose(
                    sklearn_l2.coef_,
                    skfda_l2.coef_[0],
                )

                np.testing.assert_allclose(
                    sklearn_l2.intercept_,
                    skfda_l2.intercept_,
                )

                np.testing.assert_allclose(
                    sklearn_y_pred,
                    skfda_y_pred,
                )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, TypeVar, overload

import sklearn.base

SelfType = TypeVar("SelfType")
Input = TypeVar("Input")
Output = TypeVar("Output")
Target = TypeVar("Target")


class BaseEstimator(
    ABC,
    sklearn.base.BaseEstimator,  # type: ignore[misc]
):
    pass


class TransformerMixin(
    ABC,
    Generic[Input, Output, Target],
    sklearn.base.TransformerMixin,  # type: ignore[misc]
):

    def fit(
        self: SelfType,
        X: Input,
        y: Optional[Target] = None,
    ) -> SelfType:

        return self

    @overload  # type: ignore[misc]
    def fit_transform(
        self,
        X: Input,
        y: Optional[Target] = None,
    ) -> Output:
        pass

    def fit_transform(
        self,
        X: Input,
        y: Optional[Target] = None,
        **fit_params: Any,
    ) -> Output:

        return super().fit_transform(X, y, **fit_params)


class InductiveTransformerMixin(
    TransformerMixin[Input, Output, Target],
):

    @abstractmethod
    def transform(
        self: SelfType,
        X: Input,
    ) -> Output:

        pass



"""Initialization module of visualization folder."""

from . import clustering, representation
from ._baseplot import BasePlot
from ._boxplot import Boxplot, SurfaceBoxplot
from ._ddplot import DDPlot
from ._magnitude_shape_plot import MagnitudeShapePlot
from ._multiple_display import MultipleDisplay
from ._outliergram import Outliergram
from ._parametric_plot import ParametricPlot
from .fpca import FPCAPlot



from typing import Any, Optional, Tuple, TypeVar

import numpy as np

from ..._utils import _to_domain_range
from .._typing import DomainRangeLike
from ._basis import Basis

T = TypeVar("T", bound='Fourier')


class Fourier(Basis):
    r"""Fourier basis.

    Defines a functional basis for representing functions on a fourier
    series expansion of period :math:`T`. The number of basis is always odd.
    If instantiated with an even number of basis, they will be incremented
    automatically by one.

    .. math::
        \phi_0(t) = \frac{1}{\sqrt{2}}

    .. math::
        \phi_{2n -1}(t) = sin\left(\frac{2 \pi n}{T} t\right)

    .. math::
        \phi_{2n}(t) = cos\left(\frac{2 \pi n}{T} t\right)

    Actually this basis functions are not orthogonal but not orthonormal. To
    achieve this they are divided by its norm: :math:`\sqrt{\frac{T}{2}}`.

    Parameters:
        domain_range: A tuple of length 2 containing the initial and
            end values of the interval over which the basis can be evaluated.
        n_basis: Number of functions in the basis.
        period: Period (:math:`T`).

    Examples:
        Constructs specifying number of basis, definition interval and period.

        >>> fb = Fourier((0, np.pi), n_basis=3, period=1)
        >>> fb([0, np.pi / 4, np.pi / 2, np.pi]).round(2)
        array([[[ 1.  ],
                [ 1.  ],
                [ 1.  ],
                [ 1.  ]],
               [[ 0.  ],
                [-1.38],
                [-0.61],
                [ 1.1 ]],
               [[ 1.41],
                [ 0.31],
                [-1.28],
                [ 0.89]]])

        And evaluate second derivative

        >>> deriv2 = fb.derivative(order=2)
        >>> deriv2([0, np.pi / 4, np.pi / 2, np.pi]).round(2)
        array([[[  0.  ],
                [  0.  ],
                [  0.  ],
                [  0.  ]],
               [[  0.  ],
                [ 54.46],
                [ 24.02],
                [-43.37]],
               [[-55.83],
                [-12.32],
                [ 50.4 ],
                [-35.16]]])

    """

    def __init__(
        self,
        domain_range: Optional[DomainRangeLike] = None,
        n_basis: int = 3,
        period: Optional[float] = None,
    ) -> None:
        """
        Construct a Fourier object.

        It forces the object to have an odd number of basis. If n_basis is
        even, it is incremented by one.

        Args:
            domain_range: Tuple defining the domain over which the
                function is defined.
            n_basis: Number of basis functions.
            period: Period of the trigonometric functions that
                define the basis.

        """
        if domain_range is not None:
            domain_range = _to_domain_range(domain_range)

            if len(domain_range) != 1:
                raise ValueError("Domain range should be unidimensional.")

            domain_range = domain_range[0]

        self._period = period
        # If number of basis is even, add 1
        n_basis += 1 - n_basis % 2
        super().__init__(domain_range=domain_range, n_basis=n_basis)

    @property
    def period(self) -> float:
        if self._period is None:
            return self.domain_range[0][1] - self.domain_range[0][0]

        return self._period

    def _evaluate(self, eval_points: np.ndarray) -> np.ndarray:

        # Input is scalar
        eval_points = eval_points[..., 0]

        functions = [np.sin, np.cos]
        omega = 2 * np.pi / self.period

        normalization_denominator = np.sqrt(self.period / 2)

        seq = 1 + np.arange((self.n_basis - 1) // 2)
        seq_pairs = np.array([seq, seq]).T
        phase_coefs = omega * seq_pairs

        # Multiply the phase coefficients elementwise
        res = np.einsum('ij,k->ijk', phase_coefs, eval_points)

        # Apply odd and even functions
        for i in (0, 1):
            functions[i](res[:, i, :], out=res[:, i, :])

        res = res.reshape(-1, len(eval_points))
        res /= normalization_denominator

        constant_basis = np.full(
            shape=(1, len(eval_points)),
            fill_value=1 / (np.sqrt(2) * normalization_denominator),
        )

        return np.concatenate((constant_basis, res))

    def _derivative_basis_and_coefs(
        self: T,
        coefs: np.ndarray,
        order: int = 1,
    ) -> Tuple[T, np.ndarray]:

        omega = 2 * np.pi / self.period
        deriv_factor = (np.arange(1, (self.n_basis + 1) / 2) * omega) ** order

        deriv_coefs = np.zeros(coefs.shape)

        cos_sign, sin_sign = (
            (-1) ** int((order + 1) / 2),
            (-1) ** int(order / 2),
        )

        if order % 2 == 0:
            deriv_coefs[:, 1::2] = sin_sign * coefs[:, 1::2] * deriv_factor
            deriv_coefs[:, 2::2] = cos_sign * coefs[:, 2::2] * deriv_factor
        else:
            deriv_coefs[:, 2::2] = sin_sign * coefs[:, 1::2] * deriv_factor
            deriv_coefs[:, 1::2] = cos_sign * coefs[:, 2::2] * deriv_factor

        # normalise
        return self.copy(), deriv_coefs

    def _gram_matrix(self) -> np.ndarray:

        # Orthogonal in this case
        if self.period == (self.domain_range[0][1] - self.domain_range[0][0]):
            return np.identity(self.n_basis)

        return super()._gram_matrix()

    def rescale(  # noqa: D102
        self: T,
        domain_range: Optional[DomainRangeLike] = None,
        *,
        rescale_period: bool = False,
    ) -> T:

        rescale_basis = super().rescale(domain_range)

        if rescale_period is True:

            domain_rescaled = rescale_basis.domain_range[0]
            domain = self.domain_range[0]

            rescale_basis._period = (  # noqa: WPS437
                self.period
                * (domain_rescaled[1] - domain_rescaled[0])
                / (domain[1] - domain[0])
            )

        return rescale_basis

    def _to_R(self) -> str:  # noqa: N802
        drange = self.domain_range[0]
        rangeval = f"c({drange[0]}, {drange[1]})"
        return (
            f"create.fourier.basis("
            f"rangeval = {rangeval}, "
            f"nbasis = {self.n_basis}, "
            f"period = {self.period})"
        )

    def __repr__(self) -> str:
        """Representation of a Fourier basis."""
        return (
            f"{self.__class__.__name__}("
            f"domain_range={self.domain_range}, "
            f"n_basis={self.n_basis}, "
            f"period={self.period})"
        )

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and self.period == other.period

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.period))



from __future__ import annotations

from typing import NamedTuple, Optional, Tuple

import numpy as np
import scipy.integrate
import scipy.stats
from numpy import linalg as la
from sklearn.base import BaseEstimator, OutlierMixin
from sklearn.covariance import MinCovDet

from skfda.exploratory.depth.multivariate import Depth, ProjectionDepth

from ... import FDataGrid
from ..._utils import RandomStateLike
from ...representation._typing import NDArrayFloat, NDArrayInt


class DirectionalOutlyingnessStats(NamedTuple):
    directional_outlyingness: NDArrayFloat
    functional_directional_outlyingness: NDArrayFloat
    mean_directional_outlyingness: NDArrayFloat
    variation_directional_outlyingness: NDArrayFloat


def directional_outlyingness_stats(
    fdatagrid: FDataGrid,
    *,
    multivariate_depth: Optional[Depth[NDArrayFloat]] = None,
    pointwise_weights: Optional[NDArrayFloat] = None,
) -> DirectionalOutlyingnessStats:
    r"""
    Compute the directional outlyingness of the functional data.

    Furthermore, it calculates functional, mean and the variational
    directional outlyingness of the samples in the data set, which are also
    returned.

    The functional directional outlyingness can be seen as the overall
    outlyingness, analog to other functional outlyingness measures.

    The mean directional outlyingness, describes the relative
    position (including both distance and direction) of the samples on average
    to the center curve; its norm can be regarded as the magnitude
    outlyingness.

    The variation of the directional outlyingness, measures
    the change of the directional outlyingness in terms of both norm and
    direction across the whole design interval and can be regarded as the
    shape outlyingness.

    Firstly, the directional outlyingness is calculated as follows:

    .. math::
        \mathbf{O}\left(\mathbf{X}(t) , F_{\mathbf{X}(t)}\right) =
        \left\{\frac{1}{d\left(\mathbf{X}(t) , F_{\mathbf{X}(t)}\right)} - 1
        \right\} \cdot \mathbf{v}(t)

    where :math:`\mathbf{X}` is a stochastic process with probability
    distribution :math:`F`, :math:`d` a depth function and :math:`\mathbf{v}(t)
    = \left\{ \mathbf{X}(t) - \mathbf{Z}(t)\right\} / \lVert \mathbf{X}(t) -
    \mathbf{Z}(t) \rVert` is the spatial sign of :math:`\left\{\mathbf{X}(t) -
    \mathbf{Z}(t)\right\}`, :math:`\mathbf{Z}(t)` denotes the median and
    :math:`\lVert \cdot \rVert` denotes the :math:`L_2` norm.

    From the above formula, we define the mean directional outlyingness as:

    .. math::
        \mathbf{MO}\left(\mathbf{X} , F_{\mathbf{X}}\right) = \int_I
        \mathbf{O}\left(\mathbf{X}(t) , F_{\mathbf{X}(t)}\right) \cdot w(t) dt;

    and the variation of the directional outlyingness as:

    .. math::
        VO\left(\mathbf{X} , F_{\mathbf{X}}\right) = \int_I  \lVert\mathbf{O}
        \left(\mathbf{X}(t), F_{\mathbf{X}(t)}\right)-\mathbf{MO}\left(
        \mathbf{X} , F_{\mathbf{X}}\right)  \rVert^2 \cdot w(t) dt

    where :math:`w(t)` a weight function defined on the :term:`domain` of
    :math:`\mathbf{X}`, :math:`I`.

    Then, the total functional outlyingness can be computed using these values:

    .. math::
        FO\left(\mathbf{X} , F_{\mathbf{X}}\right) = \lVert \mathbf{MO}\left(
        \mathbf{X} , F_{\mathbf{X}}\right)\rVert^2 +  VO\left(\mathbf{X} ,
        F_{\mathbf{X}}\right) .

    Args:
        fdatagrid (FDataGrid): Object containing the samples to be ordered
            according to the directional outlyingness.
        multivariate_depth (:ref:`depth measure <depth-measures>`, optional):
            Method used to order the data. Defaults to :func:`projection
            depth <skfda.exploratory.depth.multivariate.ProjectionDepth>`.
        pointwise_weights (array_like, optional): an array containing the
            weights of each point of discretisation where values have been
            recorded. Defaults to the same weight for each of the points:
            1/len(interval).

    Returns:
        DirectionalOutlyingnessStats object.

    Example:

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>> stats = directional_outlyingness_stats(fd)
        >>> stats.directional_outlyingness
        array([[[ 1.33333333],
                [ 1.33333333],
                [ 2.33333333],
                [ 1.5       ],
                [ 1.66666667],
                [ 1.66666667]],
               [[ 0.        ],
                [ 0.        ],
                [ 0.        ],
                [ 0.        ],
                [ 0.        ],
                [ 0.        ]],
               [[-1.33333333],
                [-1.33333333],
                [-1.        ],
                [-0.5       ],
                [-0.33333333],
                [-0.33333333]],
               [[-0.66666667],
                [-0.66666667],
                [-1.        ],
                [-2.5       ],
                [-3.        ],
                [-2.33333333]]])

    >>> stats.functional_directional_outlyingness
    array([ 6.58864198,  6.4608642 ,  6.63753086,  7.40481481])

    >>> stats.mean_directional_outlyingness
    array([[ 1.66666667],
           [ 0.        ],
           [-0.8       ],
           [-1.74444444]])

    >>> stats.variation_directional_outlyingness
    array([ 0.12777778,  0.        ,  0.17666667,  0.94395062])

    References:
        Dai, Wenlin, and Genton, Marc G. "Directional outlyingness for
        multivariate functional data." Computational Statistics & Data
        Analysis 131 (2019): 50-65.

    """
    if fdatagrid.dim_domain > 1:
        raise NotImplementedError("Only support 1 dimension on the domain.")

    if multivariate_depth is None:
        multivariate_depth = ProjectionDepth()

    if (
        pointwise_weights is not None
        and (
            len(pointwise_weights) != len(fdatagrid.grid_points[0])
            or pointwise_weights.sum() != 1
        )
    ):
        raise ValueError(
            "There must be a weight in pointwise_weights for each recorded "
            "time point and altogether must integrate to 1.",
        )

    if pointwise_weights is None:
        pointwise_weights = np.ones(
            len(fdatagrid.grid_points[0]),
        ) / (
            fdatagrid.domain_range[0][1] - fdatagrid.domain_range[0][0]
        )

    depth_pointwise = multivariate_depth(fdatagrid.data_matrix)
    assert depth_pointwise.shape == fdatagrid.data_matrix.shape[:-1]

    # Obtaining the pointwise median sample Z, to calculate
    # v(t) = {X(t) − Z(t)}/|| X(t) − Z(t) ||
    median_index = np.argmax(depth_pointwise, axis=0)
    pointwise_median = fdatagrid.data_matrix[
        median_index,
        range(fdatagrid.data_matrix.shape[1]),
    ]
    assert pointwise_median.shape == fdatagrid.data_matrix.shape[1:]
    v = fdatagrid.data_matrix - pointwise_median
    assert v.shape == fdatagrid.data_matrix.shape
    v_norm = la.norm(v, axis=-1, keepdims=True)

    # To avoid ZeroDivisionError, the zeros are substituted by ones (the
    # reference implementation also does this).
    v_norm[np.where(v_norm == 0)] = 1

    v_unitary = v / v_norm

    # Calculation directinal outlyingness
    dir_outlyingness = (1 / depth_pointwise[..., np.newaxis] - 1) * v_unitary

    # Calculation mean directional outlyingness
    weighted_dir_outlyingness = (
        dir_outlyingness * pointwise_weights[:, np.newaxis]
    )
    assert weighted_dir_outlyingness.shape == dir_outlyingness.shape

    mean_dir_outlyingness = scipy.integrate.simps(
        weighted_dir_outlyingness,
        fdatagrid.grid_points[0],
        axis=1,
    )
    assert mean_dir_outlyingness.shape == (
        fdatagrid.n_samples,
        fdatagrid.dim_codomain,
    )

    # Calculation variation directional outlyingness
    norm = np.square(la.norm(
        dir_outlyingness
        - mean_dir_outlyingness[:, np.newaxis, :],
        axis=-1,
    ))
    weighted_norm = norm * pointwise_weights
    variation_dir_outlyingness = scipy.integrate.simps(
        weighted_norm,
        fdatagrid.grid_points[0],
        axis=1,
    )
    assert variation_dir_outlyingness.shape == (fdatagrid.n_samples,)

    functional_dir_outlyingness = (
        np.square(la.norm(mean_dir_outlyingness))
        + variation_dir_outlyingness
    )
    assert functional_dir_outlyingness.shape == (fdatagrid.n_samples,)

    return DirectionalOutlyingnessStats(
        directional_outlyingness=dir_outlyingness,
        functional_directional_outlyingness=functional_dir_outlyingness,
        mean_directional_outlyingness=mean_dir_outlyingness,
        variation_directional_outlyingness=variation_dir_outlyingness,
    )


class MSPlotOutlierDetector(
    BaseEstimator,  # type: ignore
    OutlierMixin,  # type: ignore
):
    r"""Outlier detector using directional outlyingness.

    Considering :math:`\mathbf{Y} = \left(\mathbf{MO}^T, VO\right)^T`, the
    outlier detection method is implemented as described below.

    First, the square robust Mahalanobis distance is calculated based on a
    sample of size :math:`h \leq fdatagrid.n_samples`:

    .. math::
        {RMD}^2\left( \mathbf{Y}, \mathbf{\tilde{Y}}^*_J\right) = \left(
        \mathbf{Y} - \mathbf{\tilde{Y}}^*_J\right)^T  {\mathbf{S}^*_J}^{-1}
        \left( \mathbf{Y} - \mathbf{\tilde{Y}}^*_J\right)

    where :math:`J` denotes the group of :math:`h` samples that minimizes the
    determinant of the corresponding covariance matrix,
    :math:`\mathbf{\tilde{Y}}^*_J = h^{-1}\sum_{i\in{J}}\mathbf{Y}_i` and
    :math:`\mathbf{S}^*_J = h^{-1}\sum_{i\in{J}}\left( \mathbf{Y}_i - \mathbf{
    \tilde{Y}}^*_J\right) \left( \mathbf{Y}_i - \mathbf{\tilde{Y}}^*_J
    \right)^T`. The sub-sample of size h controls the robustness of the method.

    Then, the tail of this distance distribution is approximated as follows:

    .. math::
        \frac{c\left(m - p\right)}{m\left(p + 1\right)}RMD^2\left(
        \mathbf{Y}, \mathbf{\tilde{Y}}^*_J\right)\sim F_{p+1, m-p}

    where :math:`p` is the dimension of the image plus one, and :math:`c` and
    :math:`m` are parameters determining the degrees of freedom of the
    :math:`F`-distribution and the scaling factor, given by empirical results
    and an asymptotic formula.

    Finally, we choose a cutoff value to determine the outliers, C ,
    as the :math:`\alpha` quantile of :math:`F_{p+1, m-p}`. We set
    :math:`\alpha = 0.993`, which is used in the classical boxplot for
    detecting outliers under a normal distribution.

    Parameters:
        multivariate_depth (:ref:`depth measure <depth-measures>`, optional):
            Method used to order the data. Defaults to :class:`projection
            depth <fda.depth_measures.multivariate.ProjectionDepth>`.
        pointwise_weights (array_like, optional): an array containing the
            weights of each points of discretisati on where values have
            been recorded.
        alpha (float, optional): Denotes the quantile to choose the cutoff
            value for detecting outliers Defaults to 0.993, which is used
            in the classical boxplot.
        assume_centered (boolean, optional): If True, the support of the
            robust location and the covariance estimates is computed, and a
            covariance estimate is recomputed from it, without centering
            the data. Useful to work with data whose mean is significantly
            equal to zero but is not exactly zero. If False, default value,
            the robust location and covariance are directly computed with
            the FastMCD algorithm without additional treatment.
        support_fraction (float, 0 < support_fraction < 1, optional): The
            proportion of points to be included in the support of the
            raw MCD estimate.
            Default is None, which implies that the minimum value of
            support_fraction will be used within the algorithm:
            [n_sample + n_features + 1] / 2
        random_state (int, RandomState instance or None, optional): If int,
            random_state is the seed used by the random number generator;
            If RandomState instance, random_state is the random number
            generator; If None, the random number generator is the
            RandomState instance used by np.random. By default, it is 0.

    Example:
        Function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`.

        >>> import skfda
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> out_detector = MSPlotOutlierDetector()
        >>> out_detector.fit_predict(fd)
        array([1, 1, 1, 1])

    References:
        Dai, Wenlin, and Genton, Marc G. "Multivariate functional data
        visualization and outlier detection." Journal of Computational
        and Graphical Statistics 27.4 (2018): 923-934.

    """

    def __init__(
        self,
        *,
        multivariate_depth: Optional[Depth[NDArrayFloat]] = None,
        pointwise_weights: Optional[NDArrayFloat] = None,
        assume_centered: bool = False,
        support_fraction: Optional[float] = None,
        num_resamples: int = 1000,
        random_state: RandomStateLike = 0,
        alpha: float = 0.993,
        _force_asymptotic: bool = False,
    ) -> None:
        self.multivariate_depth = multivariate_depth
        self.pointwise_weights = pointwise_weights
        self.assume_centered = assume_centered
        self.support_fraction = support_fraction
        self.num_resamples = num_resamples
        self.random_state = random_state
        self.alpha = alpha
        self._force_asymptotic = _force_asymptotic

    def _compute_points(self, X: FDataGrid) -> NDArrayFloat:
        multivariate_depth = self.multivariate_depth
        if multivariate_depth is None:
            multivariate_depth = ProjectionDepth()

        # The depths of the samples are calculated giving them an ordering.
        *_, mean_dir_outl, variation_dir_outl = directional_outlyingness_stats(
            X,
            multivariate_depth=multivariate_depth,
            pointwise_weights=self.pointwise_weights)

        points = np.concatenate(
            (mean_dir_outl, variation_dir_outl[:, np.newaxis]),
            axis=1,
        )

        return points

    def _parameters_asymptotic(
        self,
        sample_size: int,
        dimension: int,
    ) -> Tuple[float, float]:
        """Return the scaling and cutoff parameters via asymptotic formula."""
        n = sample_size
        p = dimension

        h = np.floor((n + p + 1) / 2)

        # c estimation
        xi_left = scipy.stats.chi2.rvs(
            size=self.num_resamples,
            df=p + 2,
            random_state=self.random_state_,
        )
        xi_right = scipy.stats.ncx2.rvs(
            size=self.num_resamples,
            df=p,
            nc=h / n,
            random_state=self.random_state_,
        )

        c_numerator = np.sum(xi_left < xi_right) / self.num_resamples
        c_denominator = h / n

        estimated_c = c_numerator / c_denominator

        # m estimation
        alpha = (n - h) / n
        q_alpha = scipy.stats.chi2.ppf(1 - alpha, df=p)

        dist_p2 = scipy.stats.chi2.cdf(q_alpha, df=p + 2)
        dist_p4 = scipy.stats.chi2.cdf(q_alpha, df=p + 4)
        c_alpha = (1 - alpha) / dist_p2
        c2 = -dist_p2 / 2
        c3 = -dist_p4 / 2
        c4 = 3 * c3

        b1 = (c_alpha * (c3 - c4)) / (1 - alpha)
        b2 = (
            0.5 + c_alpha / (1 - alpha)
            * (c3 - q_alpha / p * (c2 + (1 - alpha) / 2))
        )

        v1 = (
            (1 - alpha) * b1**2
            * (alpha * (c_alpha * q_alpha / p - 1) ** 2 - 1)
            - 2 * c3 * c_alpha**2
            * (
                3 * (b1 - p * b2)**2
                + (p + 2) * b2 * (2 * b1 - p * b2)
            )
        )
        v2 = n * (b1 * (b1 - p * b2) * (1 - alpha))**2 * c_alpha**2
        v = v1 / v2

        m_asympt = 2 / (c_alpha**2 * v)

        estimated_m = (
            m_asympt * np.exp(0.725 - 0.00663 * p - 0.0780 * np.log(n))
        )

        dfn = p
        dfd = estimated_m - p + 1

        # Calculation of the cutoff value and scaling factor to identify
        # outliers.
        scaling = estimated_c * dfd / estimated_m / dfn
        cutoff_value = scipy.stats.f.ppf(self.alpha, dfn, dfd, loc=0, scale=1)

        return scaling, cutoff_value

    def _parameters_numeric(
        self,
        sample_size: int,
        dimension: int,
    ) -> Tuple[float, float]:
        from . import \
            _directional_outlyingness_experiment_results as experiments

        key = sample_size // 5

        use_asympt = True

        if dimension == 2:
            scaling_list = experiments.dim2_scaling_list
            cutoff_list = experiments.dim2_cutoff_list
            assert len(scaling_list) == len(cutoff_list)
            if key < len(scaling_list):
                use_asympt = False

        elif dimension == 3:
            scaling_list = experiments.dim3_scaling_list
            cutoff_list = experiments.dim3_cutoff_list
            assert len(scaling_list) == len(cutoff_list)
            if key < len(scaling_list):
                use_asympt = False

        if use_asympt:
            return self._parameters_asymptotic(sample_size, dimension)

        return scaling_list[key], cutoff_list[key]

    def fit_predict(self, X: FDataGrid, y: None = None) -> NDArrayInt:

        try:
            self.random_state_ = np.random.RandomState(self.random_state)
        except ValueError:
            self.random_state_ = self.random_state

        self.points_ = self._compute_points(X)

        # The square mahalanobis distances of the samples are
        # calulated using MCD.
        self.cov_ = MinCovDet(
            store_precision=False,
            assume_centered=self.assume_centered,
            support_fraction=self.support_fraction,
            random_state=self.random_state_,
        ).fit(self.points_)

        # Calculation of the degrees of freedom of the F-distribution
        # (approximation of the tail of the distance distribution).

        # One per dimension (mean dir out) plus one (variational dir out)
        dimension = X.dim_codomain + 1
        if self._force_asymptotic:
            scaling, cutoff_value = self._parameters_asymptotic(
                sample_size=X.n_samples,
                dimension=dimension,
            )
        else:
            scaling, cutoff_value = self._parameters_numeric(
                sample_size=X.n_samples,
                dimension=dimension,
            )

        self.scaling_ = scaling
        self.cutoff_value_ = cutoff_value

        rmd_2 = self.cov_.mahalanobis(self.points_)

        outliers = self.scaling_ * rmd_2 > self.cutoff_value_

        # Predict as scikit-learn outlier detectors
        predicted = ~outliers + outliers * -1

        return predicted



"""Module with generic methods."""

from __future__ import annotations

import functools
import numbers
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import scipy.integrate
from numpy import ndarray
from pandas.api.indexers import check_array_indexer
from sklearn.base import clone
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.multiclass import check_classification_targets
from typing_extensions import Literal, Protocol

from .._utils._sklearn_adapter import TransformerMixin
from ..representation._typing import (
    ArrayLike,
    DomainRange,
    DomainRangeLike,
    GridPoints,
    GridPointsLike,
    NDArrayAny,
    NDArrayFloat,
    NDArrayInt,
)
from ..representation.extrapolation import ExtrapolationLike

RandomStateLike = Optional[Union[int, np.random.RandomState]]

if TYPE_CHECKING:
    from ..exploratory.depth import Depth
    from ..representation import FData, FDataGrid
    from ..representation.basis import Basis
    T = TypeVar("T", bound=FData)

    Input = TypeVar("Input", bound=Union[FData, NDArrayFloat])
    Output = TypeVar("Output", bound=Union[FData, NDArrayFloat])
    Target = TypeVar("Target", bound=NDArrayInt)


def check_is_univariate(fd: FData) -> None:
    """Check if an FData is univariate and raises an error.

    Args:
        fd: Functional object to check if is univariate.

    Raises:
        ValueError: If it is not univariate, i.e., `fd.dim_domain != 1` or
            `fd.dim_codomain != 1`.

    """
    if fd.dim_domain != 1 or fd.dim_codomain != 1:
        domain_str = (
            "" if fd.dim_domain == 1
            else f"(currently is {fd.dim_domain}) "
        )

        codomain_str = (
            "" if fd.dim_codomain == 1
            else f"(currently is  {fd.dim_codomain})"
        )

        raise ValueError(
            f"The functional data must be univariate, i.e., "
            f"with dim_domain=1 {domain_str}"
            f"and dim_codomain=1 {codomain_str}",
        )


def _check_compatible_fdata(fdata1: FData, fdata2: FData) -> None:
    """Check that two FData are compatible."""
    if (fdata1.dim_domain != fdata2.dim_domain):
        raise ValueError(
            f"Functional data has incompatible domain dimensions: "
            f"{fdata1.dim_domain} != {fdata2.dim_domain}",
        )

    if (fdata1.dim_codomain != fdata2.dim_codomain):
        raise ValueError(
            f"Functional data has incompatible codomain dimensions: "
            f"{fdata1.dim_codomain} != {fdata2.dim_codomain}",
        )


def _check_compatible_fdatagrid(fdata1: FDataGrid, fdata2: FDataGrid) -> None:
    """Check that two FDataGrid are compatible."""
    _check_compatible_fdata(fdata1, fdata2)
    if not all(
        np.array_equal(g1, g2)
        for g1, g2 in zip(fdata1.grid_points, fdata2.grid_points)
    ):
        raise ValueError(
            f"Incompatible grid points between template and "
            f"data: {fdata1.grid_points} != {fdata2.grid_points}",
        )


def _to_grid(
    X: FData,
    y: FData,
    eval_points: Optional[NDArrayFloat] = None,
) -> Tuple[FDataGrid, FDataGrid]:
    """Transform a pair of FDatas in grids to perform calculations."""
    from .. import FDataGrid
    x_is_grid = isinstance(X, FDataGrid)
    y_is_grid = isinstance(y, FDataGrid)

    if eval_points is not None:
        X = X.to_grid(eval_points)
        y = y.to_grid(eval_points)
    elif x_is_grid and not y_is_grid:
        y = y.to_grid(X.grid_points[0])
    elif not x_is_grid and y_is_grid:
        X = X.to_grid(y.grid_points[0])
    elif not x_is_grid and not y_is_grid:
        X = X.to_grid()
        y = y.to_grid()

    return X, y


def _to_grid_points(grid_points_like: GridPointsLike) -> GridPoints:
    """Convert to grid points.

    If the original list is one-dimensional (e.g. [1, 2, 3]), return list to
    array (in this case [array([1, 2, 3])]).

    If the original list is two-dimensional (e.g. [[1, 2, 3], [4, 5]]), return
    a list containing other one-dimensional arrays (in this case
    [array([1, 2, 3]), array([4, 5])]).

    In any other case the behaviour is unespecified.

    """
    unidimensional = False

    if not isinstance(grid_points_like, Iterable):
        grid_points_like = [grid_points_like]

    if not isinstance(grid_points_like[0], Iterable):
        unidimensional = True

    if unidimensional:
        return (_int_to_real(np.asarray(grid_points_like)),)

    return tuple(_int_to_real(np.asarray(i)) for i in grid_points_like)


def _to_domain_range(sequence: DomainRangeLike) -> DomainRange:
    """Convert sequence to a proper domain range."""
    seq_aux = cast(
        Sequence[Sequence[float]],
        (sequence,) if isinstance(sequence[0], numbers.Real) else sequence,
    )

    tuple_aux = tuple(tuple(s) for s in seq_aux)

    if not all(len(s) == 2 and s[0] <= s[1] for s in tuple_aux):
        raise ValueError(
            "Domain intervals should have 2 bounds for "
            "dimension: (lower, upper).",
        )

    return cast(DomainRange, tuple_aux)


def _to_array_maybe_ragged(
    array: Iterable[ArrayLike],
    *,
    row_shape: Optional[Sequence[int]] = None,
) -> np.ndarray:
    """
    Convert to an array where each element may or may not be of equal length.

    If each element is of equal length the array is multidimensional.
    Otherwise it is a ragged array.

    """
    def convert_row(row: ArrayLike) -> np.ndarray:
        r = np.array(row)

        if row_shape is not None:
            r = r.reshape(row_shape)

        return r

    array_list = [convert_row(a) for a in array]
    shapes = [a.shape for a in array_list]

    if all(s == shapes[0] for s in shapes):
        return np.array(array_list)

    res = np.empty(len(array_list), dtype=np.object_)

    for i, a in enumerate(array_list):
        res[i] = a

    return res


@overload
def _cartesian_product(
    axes: Sequence[np.ndarray],
    *,
    flatten: bool = True,
    return_shape: Literal[False] = False,
) -> np.ndarray:
    pass


@overload
def _cartesian_product(
    axes: Sequence[np.ndarray],
    *,
    flatten: bool = True,
    return_shape: Literal[True],
) -> Tuple[np.ndarray, Tuple[int, ...]]:
    pass


def _cartesian_product(  # noqa: WPS234
    axes: Sequence[np.ndarray],
    *,
    flatten: bool = True,
    return_shape: bool = False,
) -> Union[np.ndarray, Tuple[np.ndarray, Tuple[int, ...]]]:
    """
    Compute the cartesian product of the axes.

    Computes the cartesian product of the axes and returns a numpy array of
    1 dimension with all the possible combinations, for an arbitrary number of
    dimensions.

    Args:
        axes: List with axes.
        flatten: Whether to return the flatten array or keep one dimension per
            axis.
        return_shape: If ``True`` return the shape of the array before
            flattening.

    Returns:
        Numpy 2-D array with all the possible combinations.
        The entry (i,j) represent the j-th coordinate of the i-th point.
        If ``return_shape`` is ``True`` returns also the shape of the array
        before flattening.

    Examples:
        >>> from skfda._utils import _cartesian_product
        >>> axes = [[0,1],[2,3]]
        >>> _cartesian_product(axes)
        array([[0, 2],
               [0, 3],
               [1, 2],
               [1, 3]])

        >>> axes = [[0,1],[2,3],[4]]
        >>> _cartesian_product(axes)
        array([[0, 2, 4],
               [0, 3, 4],
               [1, 2, 4],
               [1, 3, 4]])

        >>> axes = [[0,1]]
        >>> _cartesian_product(axes)
        array([[0],
               [1]])
    """
    cartesian = np.stack(np.meshgrid(*axes, indexing='ij'), -1)

    shape = cartesian.shape

    if flatten:
        cartesian = cartesian.reshape(-1, len(axes))

    if return_shape:
        return cartesian, shape

    return cartesian


def _same_domain(fd: Union[Basis, FData], fd2: Union[Basis, FData]) -> bool:
    """Check if the domain range of two objects is the same."""
    return np.array_equal(fd.domain_range, fd2.domain_range)


@overload
def _reshape_eval_points(
    eval_points: ArrayLike,
    *,
    aligned: Literal[True],
    n_samples: int,
    dim_domain: int,
) -> NDArrayFloat:
    pass


@overload
def _reshape_eval_points(
    eval_points: Sequence[ArrayLike],
    *,
    aligned: Literal[True],
    n_samples: int,
    dim_domain: int,
) -> NDArrayFloat:
    pass


@overload
def _reshape_eval_points(
    eval_points: Union[ArrayLike, Sequence[ArrayLike]],
    *,
    aligned: bool,
    n_samples: int,
    dim_domain: int,
) -> NDArrayFloat:
    pass


def _reshape_eval_points(
    eval_points: Union[ArrayLike, Iterable[ArrayLike]],
    *,
    aligned: bool,
    n_samples: int,
    dim_domain: int,
) -> NDArrayFloat:
    """Convert and reshape the eval_points to ndarray.

    Args:
        eval_points: Evaluation points to be reshaped.
        aligned: Boolean flag. True if all the samples
            will be evaluated at the same evaluation_points.
        n_samples: Number of observations.
        dim_domain: Dimension of the domain.

    Returns:
        Numpy array with the eval_points, if
        evaluation_aligned is True with shape `number of evaluation points`
        x `dim_domain`. If the points are not aligned the shape of the
        points will be `n_samples` x `number of evaluation points`
        x `dim_domain`.

    """
    if aligned:
        eval_points = np.asarray(eval_points)
    else:
        eval_points = cast(Iterable[ArrayLike], eval_points)

        eval_points = _to_array_maybe_ragged(
            eval_points,
            row_shape=(-1, dim_domain),
        )

    # Case evaluation of a single value, i.e., f(0)
    # Only allowed for aligned evaluation
    if aligned and (
        eval_points.shape == (dim_domain,)
        or (eval_points.ndim == 0 and dim_domain == 1)
    ):
        eval_points = np.array([eval_points])

    if aligned:  # Samples evaluated at same eval points
        eval_points = eval_points.reshape(
            (eval_points.shape[0], dim_domain),
        )

    else:  # Different eval_points for each sample

        if eval_points.shape[0] != n_samples:
            raise ValueError(
                f"eval_points should be a list "
                f"of length {n_samples} with the "
                f"evaluation points for each sample.",
            )

    return eval_points


def _one_grid_to_points(
    axes: GridPointsLike,
    *,
    dim_domain: int,
) -> Tuple[NDArrayFloat, Tuple[int, ...]]:
    """
    Convert a list of ndarrays, one per domain dimension, in the points.

    Returns also the shape containing the information of how each point
    is formed.
    """
    axes = _to_grid_points(axes)

    if len(axes) != dim_domain:
        raise ValueError(
            f"Length of axes should be {dim_domain}",
        )

    cartesian, shape = _cartesian_product(axes, return_shape=True)

    # Drop domain size dimension, as it is not needed to reshape the output
    shape = shape[:-1]

    return cartesian, shape


class EvaluateMethod(Protocol):
    """Evaluation method."""

    def __call__(
        self,
        __eval_points: NDArrayFloat,  # noqa: WPS112
        extrapolation: Optional[ExtrapolationLike],
        aligned: bool,
    ) -> NDArrayFloat:
        """Evaluate a function."""
        pass


@overload
def _evaluate_grid(
    axes: GridPointsLike,
    *,
    evaluate_method: EvaluateMethod,
    n_samples: int,
    dim_domain: int,
    dim_codomain: int,
    extrapolation: Optional[ExtrapolationLike] = None,
    aligned: Literal[True] = True,
) -> NDArrayFloat:
    pass


@overload
def _evaluate_grid(
    axes: Iterable[GridPointsLike],
    *,
    evaluate_method: EvaluateMethod,
    n_samples: int,
    dim_domain: int,
    dim_codomain: int,
    extrapolation: Optional[ExtrapolationLike] = None,
    aligned: Literal[False],
) -> NDArrayFloat:
    pass


def _evaluate_grid(  # noqa: WPS234
    axes: Union[GridPointsLike, Iterable[GridPointsLike]],
    *,
    evaluate_method: EvaluateMethod,
    n_samples: int,
    dim_domain: int,
    dim_codomain: int,
    extrapolation: Optional[ExtrapolationLike] = None,
    aligned: bool = True,
) -> NDArrayFloat:
    """
    Evaluate the functional object in the cartesian grid.

    This method is called internally by :meth:`evaluate` when the argument
    `grid` is True.

    Evaluates the functional object in the grid generated by the cartesian
    product of the axes. The length of the list of axes should be equal
    than the domain dimension of the object.

    If the list of axes has lengths :math:`n_1, n_2, ..., n_m`, where
    :math:`m` is equal than the dimension of the domain, the result of the
    evaluation in the grid will be a matrix with :math:`m+1` dimensions and
    shape :math:`n_{samples} x n_1 x n_2 x ... x n_m`.

    If `aligned` is false each sample is evaluated in a
    different grid, and the list of axes should contain a list of axes for
    each sample.

    If the domain dimension is 1, the result of the behaviour of the
    evaluation will be the same than :meth:`evaluate` without the grid
    option, but with worst performance.

    Args:
        axes: List of axes to generated the grid where the
            object will be evaluated.
        evaluate_method: Function used to evaluate the functional object.
        n_samples: Number of samples.
        dim_domain: Domain dimension.
        dim_codomain: Codomain dimension.
        extrapolation: Controls the
            extrapolation mode for elements outside the domain range. By
            default it is used the mode defined during the instance of the
            object.
        aligned: If False evaluates each sample
            in a different grid.
        evaluate_method: method to use to evaluate the points
        n_samples: number of samples
        dim_domain: dimension of the domain
        dim_codomain: dimensions of the codomain

    Returns:
        Numpy array with dim_domain + 1 dimensions with
            the result of the evaluation.

    Raises:
        ValueError: If there are a different number of axes than the domain
            dimension.

    """
    # Compute intersection points and resulting shapes
    if aligned:

        axes = cast(GridPointsLike, axes)

        eval_points, shape = _one_grid_to_points(axes, dim_domain=dim_domain)

    else:

        axes_per_sample = cast(Iterable[GridPointsLike], axes)

        axes_per_sample = list(axes_per_sample)

        eval_points_tuple, shape_tuple = zip(
            *[
                _one_grid_to_points(a, dim_domain=dim_domain)
                for a in axes_per_sample
            ],
        )

        if len(eval_points_tuple) != n_samples:
            raise ValueError(
                "Should be provided a list of axis per sample",
            )

        eval_points = _to_array_maybe_ragged(eval_points_tuple)

    # Evaluate the points
    evaluated = evaluate_method(
        eval_points,
        extrapolation=extrapolation,
        aligned=aligned,
    )

    # Reshape the result
    if aligned:

        res = evaluated.reshape(
            [n_samples] + list(shape) + [dim_codomain],
        )

    else:

        res = _to_array_maybe_ragged([
            r.reshape(list(s) + [dim_codomain])
            for r, s in zip(evaluated, shape_tuple)
        ])

    return res


def nquad_vec(
    func: Callable[[NDArrayFloat], NDArrayFloat],
    ranges: Sequence[Tuple[float, float]],
) -> NDArrayFloat:
    """Perform multiple integration of vector valued functions."""
    initial_depth = len(ranges) - 1

    def integrate(*args: Any, depth: int) -> NDArrayFloat:  # noqa: WPS430

        if depth == 0:
            f = functools.partial(func, *args)
        else:
            f = functools.partial(integrate, *args, depth=depth - 1)

        return scipy.integrate.quad_vec(f, *ranges[initial_depth - depth])[0]

    return integrate(depth=initial_depth)


ArrayT = TypeVar("ArrayT", bound=NDArrayAny)


def _map_in_batches(
    function: Callable[..., ArrayT],
    arguments: Tuple[Union[FData, NDArrayAny], ...],
    indexes: Tuple[NDArrayInt, ...],
    memory_per_batch: Optional[int] = None,
    **kwargs: Any,
) -> ArrayT:
    """
    Map a function over samples of FData or ndarray tuples efficiently.

    This function prevents a large set of indexes to use all available
    memory and hang the PC.

    """
    if memory_per_batch is None:
        # 256MB is not too big
        memory_per_batch = 256 * 1024 * 1024  # noqa: WPS432

    memory_per_element = sum(a.nbytes // len(a) for a in arguments)
    n_elements_per_batch_allowed = memory_per_batch // memory_per_element
    if n_elements_per_batch_allowed < 1:
        raise ValueError("Too few memory allowed for the operation")

    n_indexes = len(indexes[0])

    assert all(n_indexes == len(i) for i in indexes)

    batches: List[ArrayT] = []

    for pos in range(0, n_indexes, n_elements_per_batch_allowed):
        batch_args = tuple(
            a[i[pos:pos + n_elements_per_batch_allowed]]
            for a, i in zip(arguments, indexes)
        )

        batches.append(function(*batch_args, **kwargs))

    return np.concatenate(batches, axis=0)


def _pairwise_symmetric(
    function: Callable[..., ArrayT],
    arg1: Union[FData, NDArrayAny],
    arg2: Optional[Union[FData, NDArrayAny]] = None,
    memory_per_batch: Optional[int] = None,
    **kwargs: Any,
) -> ArrayT:
    """Compute pairwise a commutative function."""
    dim1 = len(arg1)
    if arg2 is None or arg2 is arg1:
        indices = np.triu_indices(dim1)

        matrix = np.empty((dim1, dim1))

        triang_vec = _map_in_batches(
            function,
            (arg1, arg1),
            indices,
            memory_per_batch=memory_per_batch,
            **kwargs,
        )

        # Set upper matrix
        matrix[indices] = triang_vec

        # Set lower matrix
        matrix[(indices[1], indices[0])] = triang_vec

        return matrix

    dim2 = len(arg2)
    indices = np.indices((dim1, dim2))

    vec = _map_in_batches(
        function,
        (arg1, arg2),
        (indices[0].ravel(), indices[1].ravel()),
        memory_per_batch=memory_per_batch,
        **kwargs,
    )

    return vec.reshape((dim1, dim2))


def _int_to_real(array: Union[NDArrayInt, NDArrayFloat]) -> NDArrayFloat:
    """Convert integer arrays to floating point."""
    return array + 0.0


def _check_array_key(array: NDArrayAny, key: Any) -> Any:
    """Check a getitem key."""
    key = check_array_indexer(array, key)
    if isinstance(key, tuple):
        non_ellipsis = [i for i in key if i is not Ellipsis]
        if len(non_ellipsis) > 1:
            raise KeyError(key)
        key = non_ellipsis[0]

    if isinstance(key, numbers.Integral):  # To accept also numpy ints
        key = int(key)
        if key < 0:
            key = len(array) + key

        if not 0 <= key < len(array):
            raise IndexError("index out of bounds")

        return slice(key, key + 1)

    return key


def _check_estimator(estimator):
    from sklearn.utils.estimator_checks import (
        check_get_params_invariance,
        check_set_params,
    )

    name = estimator.__name__
    instance = estimator()
    check_get_params_invariance(name, instance)
    check_set_params(name, instance)


def _classifier_get_classes(y: ndarray) -> Tuple[ndarray, NDArrayInt]:

    check_classification_targets(y)

    le = LabelEncoder()
    y_ind = le.fit_transform(y)

    classes = le.classes_

    if classes.size < 2:
        raise ValueError(
            f'The number of classes has to be greater than'
            f'one; got {classes.size} class',
        )
    return classes, y_ind


def _classifier_get_depth_methods(
    classes: ndarray,
    X: T,
    y_ind: ndarray,
    depth_methods: Sequence[Depth[T]],
) -> Sequence[Depth[T]]:
    return [
        clone(depth_method).fit(X[y_ind == cur_class])
        for cur_class in range(classes.size)
        for depth_method in depth_methods
    ]


def _classifier_fit_depth_methods(
    X: T,
    y: ndarray,
    depth_methods: Sequence[Depth[T]],
) -> Tuple[ndarray, Sequence[Depth[T]]]:
    classes, y_ind = _classifier_get_classes(y)

    class_depth_methods_ = _classifier_get_depth_methods(
        classes, X, y_ind, depth_methods,
    )

    return classes, class_depth_methods_


def _fit_feature_transformer(  # noqa: WPS320 WPS234
    X: Input,
    y: Target,
    transformer: TransformerMixin[Input, Output, Target],
) -> Tuple[
    Union[NDArrayInt, NDArrayFloat],
    Sequence[TransformerMixin[Input, Output, Target]],
]:

    classes, y_ind = _classifier_get_classes(y)

    class_feature_transformers = [
        clone(transformer).fit(X[y_ind == cur_class], y[y_ind == cur_class])
        for cur_class, _ in enumerate(classes)
    ]

    return classes, class_feature_transformers


_DependenceMeasure = Callable[[np.ndarray, np.ndarray], np.ndarray]


def _compute_dependence(
    X: Union[NDArrayInt, NDArrayFloat],
    y: Union[NDArrayInt, NDArrayFloat],
    *,
    dependence_measure: _DependenceMeasure,
) -> NDArrayFloat:
    """
    Compute dependence between points and target.

    Computes the dependence of each point in each trajectory in X with the
    corresponding class label in Y.

    """
    from dcor import rowwise

    # Move n_samples to the end
    # The shape is now input_shape + n_samples + n_output
    X = np.moveaxis(X, 0, -2)

    input_shape = X.shape[:-2]

    # Join input in a list for rowwise
    X = X.reshape(-1, X.shape[-2], X.shape[-1])

    if y.ndim == 1:
        y = np.atleast_2d(y).T
    Y = np.array([y] * len(X))

    dependence_results = rowwise(dependence_measure, X, Y)

    return dependence_results.reshape(input_shape)



"""Depth-based models for supervised classification."""
from __future__ import annotations

from itertools import combinations
from typing import Generic, Optional, Sequence, TypeVar, Union

import numpy as np
from scipy.interpolate import lagrange
from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.metrics import accuracy_score
from sklearn.pipeline import make_pipeline
from sklearn.utils.validation import check_is_fitted as sklearn_check_is_fitted

from ..._utils import _classifier_fit_depth_methods, _classifier_get_classes
from ...exploratory.depth import Depth, ModifiedBandDepth
from ...preprocessing.dim_reduction.feature_extraction import DDGTransformer
from ...representation._typing import NDArrayFloat, NDArrayInt
from ...representation.grid import FData

T = TypeVar("T", bound=FData)


class DDClassifier(
    BaseEstimator,  # type: ignore
    ClassifierMixin,  # type: ignore
    Generic[T],
):
    """Depth-versus-depth (DD) classifer for functional data.

    Transforms the data into a DD-plot and then classifies using a polynomial
    of a chosen degree. The polynomial passes through zero and maximizes the
    accuracy of the classification on the train dataset.

    If a point is below the polynomial in the DD-plot, it is classified to
    the first class. Otherwise, the point is classified to the second class.

    Parameters:
        degree: degree of the polynomial used to classify in the DD-plot
        depth_method:
            The depth class to use when calculating the depth of a test
            sample in a class. See the documentation of the depths module
            for a list of available depths. By default it is ModifiedBandDepth.

    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split
        >>> dataset = fetch_growth()
        >>> fd = dataset['data']
        >>> y = dataset['target']
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     fd, y, test_size=0.25, stratify=y, random_state=0)

        We will fit a DD-classifier

        >>> from skfda.ml.classification import DDClassifier
        >>> clf = DDClassifier(degree=2)
        >>> clf.fit(X_train, y_train)
        DDClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X_test)
        array([1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
               1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])

        Finally, we calculate the mean accuracy for the test data

        >>> clf.score(X_test, y_test)
        0.875

    See also:
        :class:`~skfda.ml.classification.DDGClassifier`
        :class:`~skfda.ml.classification.MaximumDepthClassifier`
        :class:`~skfda.preprocessing.dim_reduction.feature_extraction._ddg_transformer`

    References:
        Li, J., Cuesta-Albertos, J. A., and Liu, R. Y. (2012). DD-classifier:
        Nonparametric classification procedure based on DD-plot. Journal of
        the American Statistical Association, 107(498):737-753.
    """

    def __init__(
        self,
        degree: int,
        depth_method: Optional[Depth[T]] = None,
    ) -> None:
        self.depth_method = depth_method
        self.degree = degree

    def fit(self, X: T, y: NDArrayInt) -> DDClassifier[T]:
        """Fit the model using X as training data and y as target values.

        Args:
            X: FDataGrid with the training data.
            y: Target values of shape = (n_samples).

        Returns:
            self
        """
        if self.depth_method is None:
            self.depth_method = ModifiedBandDepth()

        classes, class_depth_methods = _classifier_fit_depth_methods(
            X, y, [self.depth_method],
        )

        self._classes = classes
        self.class_depth_methods_ = class_depth_methods

        if (len(self._classes) != 2):
            raise ValueError("DDClassifier only accepts two classes.")

        dd_coordinates = [
            depth_method.transform(X)
            for depth_method in self.class_depth_methods_
        ]

        polynomial_elements = combinations(
            range(len(dd_coordinates[0])),  # noqa: WPS518
            self.degree,
        )

        accuracy = -1  # initialise accuracy

        for elements in polynomial_elements:
            x_coord = np.append(dd_coordinates[0][list(elements)], 0)
            y_coord = np.append(dd_coordinates[1][list(elements)], 0)

            poly = lagrange(
                x_coord, y_coord,
            )

            predicted_values = np.polyval(poly, dd_coordinates[0])

            y_pred = self._classes[(
                dd_coordinates[1] > predicted_values
            ).astype(int)
            ]

            new_accuracy = accuracy_score(y, y_pred)

            if (new_accuracy > accuracy):
                accuracy = new_accuracy
                self.poly_ = poly

        return self

    def predict(self, X: T) -> NDArrayInt:
        """Predict the class labels for the provided data.

        Args:
            X: FDataGrid with the test samples.

        Returns:
            Array of shape (n_samples) with class labels
                for each data sample.
        """
        sklearn_check_is_fitted(self)

        dd_coordinates = [
            depth_method.transform(X)
            for depth_method in self.class_depth_methods_
        ]

        predicted_values = np.polyval(self.poly_, dd_coordinates[0])

        return self._classes[(
            dd_coordinates[1] > predicted_values
        ).astype(int)
        ]


class DDGClassifier(
    BaseEstimator,  # type: ignore
    ClassifierMixin,  # type: ignore
    Generic[T],
):
    r"""Generalized depth-versus-depth (DD) classifer for functional data.

    This classifier builds an interface around the DDGTransfomer.

    The transformer takes a list of k depths and performs the following map:

    .. math::
        \mathcal{X} &\rightarrow \mathbb{R}^G \\
        x &\rightarrow \textbf{d} = (D_1^1(x), D_1^2(x),...,D_g^k(x))

    Where :math:`D_i^j(x)` is the depth of the point :math:`x` with respect to
    the data in the :math:`i`-th group using the :math:`j`-th depth of the
    provided list.

    Note that :math:`\mathcal{X}` is possibly multivariate, that is,
    :math:`\mathcal{X} = \mathcal{X}_1 \times ... \times \mathcal{X}_p`.

    In the G dimensional space the classification is performed using a
    multivariate classifer.

    Parameters:
        depth_method:
            The depth class or sequence of depths to use when calculating
            the depth of a test sample in a class. See the documentation of
            the depths module for a list of available depths. By default it
            is ModifiedBandDepth.
        multivariate_classifier:
            The multivariate classifier to use in the DDG-plot.

    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split
        >>> dataset = fetch_growth()
        >>> fd = dataset['data']
        >>> y = dataset['target']
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     fd, y, test_size=0.25, stratify=y, random_state=0)

        >>> from sklearn.neighbors import KNeighborsClassifier

        We will fit a DDG-classifier using KNN

        >>> from skfda.ml.classification import DDGClassifier
        >>> clf = DDGClassifier(KNeighborsClassifier())
        >>> clf.fit(X_train, y_train)
        DDGClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X_test)
        array([1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1,
               1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])

        Finally, we calculate the mean accuracy for the test data

        >>> clf.score(X_test, y_test)
        0.875

    See also:
        :class:`~skfda.ml.classification.DDClassifier`
        :class:`~skfda.ml.classification.MaximumDepthClassifier`
        :class:`~skfda.preprocessing.dim_reduction.feature_extraction._ddg_transformer`

    References:
        Li, J., Cuesta-Albertos, J. A., and Liu, R. Y. (2012). DD-classifier:
        Nonparametric classification procedure based on DD-plot. Journal of
        the American Statistical Association, 107(498):737-753.

        Cuesta-Albertos, J.A., Febrero-Bande, M. and Oviedo de la Fuente, M.
        (2017) The DDG-classifier in the functional setting. TEST, 26. 119-142.
    """

    def __init__(  # noqa: WPS234
        self,
        multivariate_classifier: ClassifierMixin = None,
        depth_method: Union[Depth[T], Sequence[Depth[T]], None] = None,
    ) -> None:
        self.multivariate_classifier = multivariate_classifier
        self.depth_method = depth_method

    def fit(self, X: T, y: NDArrayInt) -> DDGClassifier[T]:
        """Fit the model using X as training data and y as target values.

        Args:
            X: FDataGrid with the training data.
            y: Target values of shape = (n_samples).

        Returns:
            self
        """
        self._pipeline = make_pipeline(
            DDGTransformer(self.depth_method),
            clone(self.multivariate_classifier),
        )

        self._pipeline.fit(X, y)

        return self

    def predict(self, X: T) -> NDArrayInt:
        """Predict the class labels for the provided data.

        Args:
            X: FDataGrid with the test samples.

        Returns:
            Array of shape (n_samples) with class labels
                for each data sample.
        """
        return self._pipeline.predict(X)


class _ArgMaxClassifier(
    BaseEstimator,  # type: ignore
    ClassifierMixin,  # type: ignore
):
    r"""Arg max classifier for multivariate data.

    Test samples are classified to the class that corresponds to the
    index of the highest coordinate.

    Examples:
        >>> import numpy as np
        >>> X = np.array([[1,5], [3,2], [4,1]])
        >>> y = np.array([1, 0, 0])

        We will fit am ArgMax classifier

        >>> from skfda.ml.classification._depth_classifiers import \
        ... _ArgMaxClassifier
        >>> clf = _ArgMaxClassifier()
        >>> clf.fit(X, y)
        _ArgMaxClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X) # Predict labels for test samples
        array([1, 0, 0])
    """

    def fit(self, X: NDArrayFloat, y: NDArrayInt) -> _ArgMaxClassifier:
        """Fit the model using X as training data and y as target values.

        Args:
            X: Array with the training data.
            y: Target values of shape = (n_samples).

        Returns:
            self
        """
        classes, _ = _classifier_get_classes(y)
        self._classes = classes
        return self

    def predict(self, X: NDArrayFloat) -> NDArrayInt:
        """Predict the class labels for the provided data.

        Args:
            X: Array with the test samples.

        Returns:
            Array of shape (n_samples) with class labels
                for each data sample.
        """
        return self._classes[np.argmax(X, axis=1)]


class MaximumDepthClassifier(DDGClassifier[T]):
    """Maximum depth classifier for functional data.

    Test samples are classified to the class where they are deeper.

    Parameters:
        depth_method:
            The depth class to use when calculating the depth of a test
            sample in a class. See the documentation of the depths module
            for a list of available depths. By default it is ModifiedBandDepth.
    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split
        >>> dataset = fetch_growth()
        >>> fd = dataset['data']
        >>> y = dataset['target']
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     fd, y, test_size=0.25, stratify=y, random_state=0)

        We will fit a Maximum depth classifier

        >>> from skfda.ml.classification import MaximumDepthClassifier
        >>> clf = MaximumDepthClassifier()
        >>> clf.fit(X_train, y_train)
        MaximumDepthClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X_test) # Predict labels for test samples
        array([1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
               1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])

        Finally, we calculate the mean accuracy for the test data

        >>> clf.score(X_test, y_test)
        0.875

    See also:
        :class:`~skfda.ml.classification.DDClassifier`
        :class:`~skfda.ml.classification.DDGClassifier`

    References:
        Ghosh, A. K. and Chaudhuri, P. (2005b). On maximum depth and
        related classifiers. Scandinavian Journal of Statistics, 32, 327–350.
    """

    def __init__(self, depth_method: Optional[Depth[T]] = None) -> None:
        super().__init__(_ArgMaxClassifier(), depth_method)
